#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to the disparity map computation step.
"""

import logging
from abc import ABCMeta, abstractmethod
from typing import Dict, Union, Tuple

import numpy as np
import xarray as xr
from json_checker import Checker, And, Or
from scipy.ndimage.morphology import binary_dilation

import pandora.constants as cst
from pandora.img_tools import compute_std_raster


class AbstractDisparity():
    """
    Abstract Disparity class
    """
    __metaclass__ = ABCMeta

    disparity_methods_avail = {}

    def __new__(cls, **cfg: dict) -> object:
        """
        Return the plugin associated with the validation_method given in the configuration

        :param cfg: configuration {'validation_method': value}
        :type cfg: dictionary
        """
        if cls is AbstractDisparity:
            if isinstance(cfg['disparity_method'], str):
                try:
                    return super(AbstractDisparity, cls).__new__(cls.disparity_methods_avail[cfg['disparity_method']])
                except KeyError:
                    logging.error('No disparity method named % supported', cfg['disparity_method'])
                    raise KeyError
            else:
                if isinstance(cfg['disparity_method'], unicode): # pylint: disable=undefined-variable
                    # creating a plugin from registered short name given as unicode (py2 & 3 compatibility)
                    try:
                        return super(AbstractDisparity, cls).__new__(
                            cls.disparity_methods_avail[cfg['disparity_method'].encode('utf-8')])
                    except KeyError:
                        logging.error(
                            'No disparity method named % supported', cfg['disparity_method'])
                        raise KeyError
        else:
            return super(AbstractDisparity, cls).__new__(cls)

    @classmethod
    def register_subclass(cls, short_name: str):
        """
        Allows to register the subclass with its short name

        :param short_name: the subclass to be registered
        :type short_name: string
        """

        def decorator(subclass):
            """
            Registers the subclass in the available methods

            :param subclass: the subclass to be registered
            :type subclass: object
            """
            cls.disparity_methods_avail[short_name] = subclass
            return subclass

        return decorator

    @abstractmethod
    def desc(self):
        """
        Describes the disparity method
        """
        print('Disparity method description')

    @abstractmethod
    def to_disp(self, cv: xr.Dataset, img_left: xr.Dataset = None, img_right: xr.Dataset = None) -> xr.Dataset:
        """
        Disparity computation by applying the Winner Takes All strategy

        :param cv: the cost volume datset with the data variables:

                - cv 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset,
        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :return: Dataset with the disparity map and the confidence measure with the data variables :

                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :rtype: xarray.Dataset
        """

    @staticmethod
    def coefficient_map(cv: xr.DataArray) -> xr.DataArray:
        """
        Return the coefficient map

        :param cv: cost volume
        :type cv: xarray.Dataset, with the data variables cost_volume 3D xarray.DataArray (row, col, disp)
        :return: the coefficient map
        :rtype: 2D DataArray (row, col)
        """
        row = cv.coords['row']
        col = cv.coords['col']

        # Create the coefficient map
        coeff_map = xr.DataArray(cv['cost_volume'].sel(disp=cv['disp_indices']).astype(np.float32),
                                 coords=[('row', row), ('col', col)])
        coeff_map.name = 'Coefficient Map'
        coeff_map.attrs = cv.attrs

        return coeff_map

    @staticmethod
    def approximate_right_disparity(cv: xr.Dataset, img_right: xr.Dataset,
                                    invalid_value: float = 0) -> xr.Dataset:
        """
        Create the right disparity map, by a diagonal search for the minimum in the left cost volume

        ERNST, Ines et HIRSCHMÜLLER, Heiko.
        Mutual information based semi-global stereo matching on the GPU.
        In : International Symposium on Visual Computing. Springer, Berlin, Heidelberg, 2008. p. 228-239.

        :param cv: the cost volume dataset with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param invalid_value: disparity to assign to invalid pixels
        :type invalid_value: float
        :return: Dataset with the right disparity map, the confidence measure and the validity mask with \
        the data variables :

                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
                - validity_mask 2D xarray.DataArray (row, col)
        :rtype: xarray.Dataset
        """
        disp_range = cv.coords['disp'].data.astype(float)
        col_range = cv.coords['col'].data
        row_range = cv.coords['row'].data
        # Extract integer disparity
        disp_range = np.extract(np.mod(disp_range, 1) == 0, disp_range)

        # Allocate the disparity map
        data = np.zeros((len(row_range), len(col_range))).astype(np.float32)
        disp_map = xr.Dataset({'disparity_map': (['row', 'col'], data)},
                              coords={'row': cv.coords['row'], 'col': cv.coords['col']})

        confidence_measure = compute_std_raster(img_right, cv.attrs['window_size'])

        # Create the confidence measure with the original image size and fill it
        confidence_measure_full = np.full((len(row_range), len(col_range), 1), np.nan, dtype=np.float32)
        offset = cv.attrs['offset_row_col']
        row_off = np.arange(row_range[0] + offset, row_range[-1] - offset + 1)
        col_off = np.arange(col_range[0] + offset, col_range[-1] - offset + 1)
        if offset != 0:
            confidence_measure_full[offset: - offset, offset: - offset, :] = confidence_measure.reshape(
                (len(row_off), len(col_off), 1))
        else:
            confidence_measure_full = confidence_measure.reshape((len(row_range), len(col_range), 1))

        disp_map['confidence_measure'] = xr.DataArray(data=confidence_measure_full.astype(np.float32),
                                                      dims=['row', 'col', 'indicator'])

        # Allocate the validity mask
        disp_map['validity_mask'] = xr.DataArray(np.zeros(disp_map['disparity_map'].shape, dtype=np.uint16),
                                                 dims=['row', 'col'])

        disp_map.attrs = cv.attrs
        d_range = cv.coords['disp'].data
        disp_map.attrs['disp_min'] = d_range[0]
        disp_map.attrs['disp_max'] = d_range[-1]
        disp_map.attrs['right_left_mode'] = 'approximate'
        offset = disp_map.attrs['offset_row_col']

        indices_nan = np.isnan(cv['cost_volume'].data)
        if cv.attrs['type_measure'] == 'max':
            cv['cost_volume'].data[indices_nan] = -np.inf
        else:
            cv['cost_volume'].data[indices_nan] = np.inf

        for col in col_range:
            x_d = col - disp_range
            valid = np.where((x_d >= col_range[0]) & (x_d <= col_range[-1]))

            # The disparity interval is missing in the left image
            if x_d[valid].size == 0:
                disp_map['disparity_map'].loc[dict(col=col)] = invalid_value

                # Invalid pixel : the disparity interval is missing in the right image
                disp_map['validity_mask'].loc[
                    dict(col=col)] += cst.PANDORA_MSK_PIXEL_RIGHT_NODATA_OR_DISPARITY_RANGE_MISSING
            else:
                # Diagonal search for the minimum or maximum

                if cv.attrs['type_measure'] == 'max':
                    min_ = cv['cost_volume'].sel(col=xr.DataArray(np.flip(x_d[valid]), dims='disp_'),
                                                 disp=xr.DataArray(np.flip(disp_range[valid]),
                                                                   dims='disp_')).argmax(dim='disp_')
                else:
                    min_ = cv['cost_volume'].sel(col=xr.DataArray(np.flip(x_d[valid]), dims='disp_'),
                                                 disp=xr.DataArray(np.flip(disp_range[valid]),
                                                                   dims='disp_')).argmin(dim='disp_')
                # Disparity interval is incomplete
                if x_d[valid].size != disp_range.size:
                    #  Information: the disparity interval is incomplete (border reached in the right image)
                    disp_map['validity_mask'].loc[dict(col=col)] += \
                        cst.PANDORA_MSK_PIXEL_RIGHT_INCOMPLETE_DISPARITY_RANGE

                disp_map['disparity_map'].loc[dict(col=col)] = -1 * np.flip(disp_range[valid])[min_.data]

        cv['cost_volume'].data[indices_nan] = np.nan
        invalid_mc = np.min(indices_nan, axis=2)
        # Pixels where the disparity interval is missing in the right image, have a disparity value invalid_value
        invalid_pixel = np.where(invalid_mc)
        disp_map['disparity_map'].data[invalid_pixel] = invalid_value

        if offset > 0:
            AbstractDisparity.mask_border(disp_map)

        return disp_map

    def validity_mask(self, disp: xr.Dataset, img_left: xr.Dataset, img_right: xr.Dataset,
                      cv: xr.Dataset) -> None:
        """
        Create the validity mask of the disparity map

        :param disp: dataset with the disparity map and the confidence measure
        :type disp: xarray.Dataset with the data variables :

                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray(row, col, indicator)
        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param cv: cost volume dataset with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :return: None
        """
        # Allocate the validity mask
        disp['validity_mask'] = xr.DataArray(np.zeros(disp['disparity_map'].shape, dtype=np.uint16),
                                             dims=['row', 'col'])

        d_min = int(disp.attrs['disp_min'])
        d_max = int(disp.attrs['disp_max'])
        col = disp.coords['col'].data

        # Since disparity map is full size (input images size)
        offset = disp.attrs['offset_row_col']

        # Negative disparity range
        if d_max < 0:
            bit_1 = np.where((col + d_max) < (col[0]+offset))
            # Information: the disparity interval is incomplete (border reached in the right image)
            disp['validity_mask'].data[:, np.where(((col + d_max) >= (col[0]+offset)) &
                                                   ((col + d_min) < (col[0]+offset)))] \
                +=cst.PANDORA_MSK_PIXEL_RIGHT_INCOMPLETE_DISPARITY_RANGE
        else:
            # Positive disparity range
            if d_min > 0:
                bit_1 = np.where((col + d_min) > (col[-1]-offset))
                # Information: the disparity interval is incomplete (border reached in the right image)
                disp['validity_mask'].data[:, np.where(((col + d_min) <= (col[-1]-offset)) &
                                                       ((col + d_max) > (col[-1]-offset)))] += \
                    cst.PANDORA_MSK_PIXEL_RIGHT_INCOMPLETE_DISPARITY_RANGE

            # Disparity range contains 0
            else:
                bit_1 = ([],)
                # Information: the disparity interval is incomplete (border reached in the right image)
                disp['validity_mask'].data[:, np.where(((col + d_min) < (col[0]+offset)) |
                                                       (col + d_max > (col[-1])-offset))] += \
                    cst.PANDORA_MSK_PIXEL_RIGHT_INCOMPLETE_DISPARITY_RANGE

        # Invalid pixel : the disparity interval is missing in the right image ( disparity range
        # outside the image )
        disp['validity_mask'].data[:, bit_1] += cst.PANDORA_MSK_PIXEL_RIGHT_NODATA_OR_DISPARITY_RANGE_MISSING

        # The disp_min and disp_max used to search missing disparity interval are not the local disp_min and disp_max
        # in case of a variable range of disparities. So there may be pixels that have missing disparity range (all
        # cost are np.nan), but are not detected in the code block above. To find the pixels that have a missing
        # disparity range, we search in the cost volume pixels where cost_volume(row,col, for all d) = np.nan

        self.mask_invalid_variable_disparity_range(disp, cv)

        if 'msk' in img_left.data_vars:
            self.allocate_left_mask(disp, img_left)

        if 'msk' in img_right.data_vars:
            self.allocate_right_mask(disp, img_right, bit_1)
        if offset > 0:
            self.mask_border(disp)

    @staticmethod
    def mask_border(disp: xr.Dataset):
        """
        Mask border pixel  which haven't been calculated because of the window's size

        :param disp: dataset with the disparity map and the confidence measure  with the data variables :

                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray(row, col, indicator)
        :type disp: xarray.Dataset
        :return: None
        """
        offset = disp.attrs['offset_row_col']

        if offset > 0:
            # Border pixels have invalid disparity, erase the potential previous values
            disp['validity_mask'].data[:offset, :] = cst.PANDORA_MSK_PIXEL_LEFT_NODATA_OR_BORDER
            disp['validity_mask'].data[-offset:, :] = cst.PANDORA_MSK_PIXEL_LEFT_NODATA_OR_BORDER
            disp['validity_mask'].data[offset:-offset, :offset] = cst.PANDORA_MSK_PIXEL_LEFT_NODATA_OR_BORDER
            disp['validity_mask'].data[offset:-offset, -offset:] = cst.PANDORA_MSK_PIXEL_LEFT_NODATA_OR_BORDER


    @staticmethod
    def mask_invalid_variable_disparity_range(disp, cv) -> None:
        """
        Mask the pixels that have a missing disparity range, searching in the cost volume
        the pixels where cost_volume(row,col, for all d) = np.nan

        :param disp: dataset with the disparity map and the confidence measure with the data variables :

                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray(row, col, indicator)
        :type disp: xarray.Dataset
        :param cv: cost volume dataset with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :return: None
        """
        indices_nan = np.isnan(cv['cost_volume'].data)
        missing_disparity_range = np.min(indices_nan, axis=2)
        missing_range_y, missing_range_x = np.where(missing_disparity_range)

        # Mask the positions which have an missing disparity range, not already taken into account
        condition_to_mask = (disp['validity_mask'].data[missing_range_y, missing_range_x] &
                             cst.PANDORA_MSK_PIXEL_RIGHT_NODATA_OR_DISPARITY_RANGE_MISSING) == 0
        masking_value = disp['validity_mask'].data[missing_range_y, missing_range_x] + \
                        cst.PANDORA_MSK_PIXEL_RIGHT_NODATA_OR_DISPARITY_RANGE_MISSING
        no_masking_value = disp['validity_mask'].data[missing_range_y, missing_range_x]

        disp['validity_mask'].data[missing_range_y, missing_range_x] = np.where(condition_to_mask,
                                                                                masking_value,
                                                                                no_masking_value)

    @staticmethod
    def allocate_left_mask(disp: xr.Dataset, img_left: xr.Dataset) -> None:
        """
        Allocate the left image mask

        :param disp: dataset with the disparity map and the confidence measure with the data variables :

                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray(row, col, indicator)
        :type disp: xarray.Dataset
        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :return: None
        """
        _, r_mask = xr.align(disp['validity_mask'], img_left['msk'])

        # Dilatation : pixels that contains no_data in their aggregation window become no_data
        dil = binary_dilation(img_left['msk'].data == img_left.attrs['no_data_mask'],
                              structure=np.ones((disp.attrs['window_size'], disp.attrs['window_size'])),
                              iterations=1)

        # Invalid pixel : no_data in the left image
        disp['validity_mask'] += dil.astype(np.uint16) * cst.PANDORA_MSK_PIXEL_LEFT_NODATA_OR_BORDER

        # Invalid pixel : invalidated by the validity mask of the left image given as input
        disp['validity_mask'] += xr.where(
            (r_mask != img_left.attrs['no_data_mask']) & (r_mask != img_left.attrs['valid_pixels']),
            cst.PANDORA_MSK_PIXEL_IN_VALIDITY_MASK_LEFT, 0).astype(np.uint16)

    @staticmethod
    def allocate_right_mask(disp: xr.Dataset, img_right: xr.Dataset, bit_1: Union[np.ndarray, Tuple]) -> None:
        """
        Allocate the right image mask

        :param disp: dataset with the disparity map and the confidence measure with the data variables :

                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray(row, col, indicator)
        :type disp: xarray.Dataset
        :param img_right: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param bit_1: where the disparity interval is missing in the right image ( disparity range outside the image )
        :type: ndarray or Tuple
        :return: None
        """

        offset = disp.attrs['offset_row_col']

        _, r_mask = xr.align(disp['validity_mask'], img_right['msk'])
        d_min = int(disp.attrs['disp_min'])
        d_max = int(disp.attrs['disp_max'])
        col = disp.coords['col'].data
        row = disp.coords['row'].data

        # Dilatation : pixels that contains no_data in their aggregation window become no_data
        dil = binary_dilation(img_right['msk'].data == img_right.attrs['no_data_mask'],
                              structure=np.ones((disp.attrs['window_size'], disp.attrs['window_size'])),
                              iterations=1)

        r_mask = xr.where((r_mask != img_right.attrs['no_data_mask']) & (r_mask != img_right.attrs['valid_pixels']),
                          1, 0).data

        # Useful to calculate the case where the disparity interval is incomplete, and all remaining right
        # positions are invalidated by the right mask
        b_2_7 = np.zeros((len(row), len(col)), dtype=np.uint16)
        # Useful to calculate the case where no_data in the right image invalidated the disparity interval
        no_data_right = np.zeros((len(row), len(col)), dtype=np.uint16)

        col_range = np.arange(len(col))
        for dsp in range(d_min, d_max + 1):
            # Diagonal in the cost volume
            col_d = col_range + dsp
            valid_index = np.where((col_d >= col_range[0]+offset) & (col_d <= col_range[-1]-offset))

            # No_data and masked pixels do not raise the same flag, we need to treat them differently
            b_2_7[:, col_range[valid_index]] += r_mask[:, col_d[valid_index]].astype(np.uint16)
            b_2_7[:, col_range[np.setdiff1d(col_range, valid_index)]] += 1

            no_data_right[:, col_range[valid_index]] += dil[:, col_d[valid_index]]
            no_data_right[:, col_range[np.setdiff1d(col_range, valid_index)]] += 1

            # Exclusion of pixels that have flag 1 already enabled
            b_2_7[:, bit_1[0]] = 0
            no_data_right[:, bit_1[0]] = 0

            # Invalid pixel: right positions invalidated by the mask of the right image given as input
            disp['validity_mask'].data[np.where(b_2_7 == len(range(d_min, d_max + 1)))] += \
                cst.PANDORA_MSK_PIXEL_IN_VALIDITY_MASK_RIGHT

            # If Invalid pixel : the disparity interval is missing in the right image (disparity interval
            # is invalidated by no_data in the right image )
            disp['validity_mask'].data[np.where(no_data_right == len(range(d_min, d_max + 1)))] += \
                cst.PANDORA_MSK_PIXEL_RIGHT_NODATA_OR_DISPARITY_RANGE_MISSING


@AbstractDisparity.register_subclass('wta')
class WinnerTakesAll(AbstractDisparity):
    """
    WinnerTakesAll class allows to perform the disparity computation step
    """
    # Default configuration, do not change this value
    _INVALID_DISPARITY = -9999

    def __init__(self, **cfg):
        """float
        :param cfg: optional configuration
        :type cfg: dictionary
        """
        self.cfg = self.check_conf(**cfg)
        self._invalid_disparity = self.cfg['invalid_disparity']

    def check_conf(self, **cfg: Union[str, int, float, bool]) -> Dict[str, Union[str, int, float, bool]]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: disparity configuration
        :type cfg: dict
        :return cfg: disparity configuration updated
        :rtype: dict
        """
        # Give the default value if the required element is not in the configuration
        if 'invalid_disparity' not in cfg:
            cfg['invalid_disparity'] = self._INVALID_DISPARITY

        schema = {
            'disparity_method': And(str, lambda input: 'wta'),
            'invalid_disparity': Or(int, float, lambda input: np.isnan(input))
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self) -> None:
        """
        Describes the disparity method
        :return: None
        """
        print('Winner takes all method')

    def to_disp(self, cv: xr.Dataset, img_left: xr.Dataset = None, img_right: xr.Dataset = None) -> xr.Dataset:
        """
        Disparity computation by applying the Winner Takes All strategy

        :param cv: the cost volume datset with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :return: Dataset with the disparity map and the confidence measure  with the data variables :

                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :rtype: xarray.Dataset
        """
        indices_nan = np.isnan(cv['cost_volume'].data)

        # Winner Takes All strategy
        if cv.attrs['type_measure'] == 'max':
            # Disparities are computed by selecting the maximal cost at each pixel
            cv['cost_volume'].data[indices_nan] = -np.inf
            disp = self.argmax_split(cv)
        else:
            # Disparities are computed by selecting the minimal cost at each pixel
            cv['cost_volume'].data[indices_nan] = np.inf
            disp = self.argmin_split(cv)

        cv['cost_volume'].data[indices_nan] = np.nan
        row = cv.coords['row']
        col = cv.coords['col']

        # ----- Disparity map -----
        disp_map = xr.Dataset({'disparity_map': (['row', 'col'], disp)}, coords={'row': row, 'col': col})
        invalid_mc = np.min(indices_nan, axis=2)
        # Pixels where the disparity interval is missing in the right image, have a disparity value invalid_value
        invalid_pixel = np.where(invalid_mc)
        disp_map['disparity_map'].data[invalid_pixel] = self._invalid_disparity

        # Save the disparity map in the cost volume
        cv['disp_indices'] = disp_map['disparity_map'].copy(deep=True)

        disp_map.attrs = cv.attrs
        d_range = cv.coords['disp'].data
        disp_map.attrs['disp_min'] = d_range[0]
        disp_map.attrs['disp_max'] = d_range[-1]

        # ----- Confidence measure -----
        # Allocate the confidence measure in the disparity_map dataset
        disp_map['confidence_measure'] = cv['confidence_measure']

        # Remove temporary values
        del indices_nan
        del invalid_mc

        return disp_map

    @staticmethod
    def argmin_split(cost_volume: xr.Dataset) -> np.ndarray:
        """
        Find the indices of the minimum values for a 3D DataArray, along axis 2.
        Memory consumption is reduced by splitting the 3D Array.

        :param cost_volume: the cost volume dataset
        :type cost_volume: xarray.Dataset
        :return: the disparities for which the cost volume values are the smallest
        :rtype: np.ndarray
        """
        ncol, nrow, ndsp = cost_volume['cost_volume'].shape # pylint: disable=unused-variable
        disp = np.zeros((ncol, nrow), dtype=np.float32)

        # Numpy argmin is making a copy of the cost volume.
        # To reduce memory, numpy argmin is applied on a small part of the cost volume.
        # The cost volume is split (along the row axis) into multiple sub-arrays with a step of 100
        cv_chunked_y = np.array_split(cost_volume['cost_volume'].data, np.arange(100, ncol, 100), axis=0)

        y_begin = 0

        for col, cv_y in enumerate(cv_chunked_y): # pylint: disable=unused-variable
            # To reduce memory, the cost volume is split (along the col axis) into
            # multiple sub-arrays with a step of 100
            cv_chunked_x = np.array_split(cv_y, np.arange(100, nrow, 100), axis=1)
            x_begin = 0
            for row, cv_x in enumerate(cv_chunked_x): # pylint: disable=unused-variable
                disp[y_begin:y_begin + cv_y.shape[0], x_begin: x_begin + cv_x.shape[1]] = \
                    cost_volume.coords['disp'].data[np.argmin(cv_x, axis=2)]
                x_begin += cv_x.shape[1]

            y_begin += cv_y.shape[0]

        return disp

    @staticmethod
    def argmax_split(cost_volume: xr.Dataset) -> np.ndarray:
        """
        Find the indices of the maximum values for a 3D DataArray, along axis 2.
        Memory consumption is reduced by splitting the 3D Array.

        :param cost_volume: the cost volume dataset
        :type cost_volume: xarray.Dataset
        :return: the disparities for which the cost volume values are the highest
        :rtype: np.ndarray
        """
        ncol, nrow, ndisp = cost_volume['cost_volume'].shape # pylint: disable=unused-variable
        disp = np.zeros((ncol, nrow), dtype=np.float32)

        # Numpy argmax is making a copy of the cost volume.
        # To reduce memory, numpy argmax is applied on a small part of the cost volume.
        # The cost volume is split (along the row axis) into multiple sub-arrays with a step of 100
        cv_chunked_col = np.array_split(cost_volume['cost_volume'].data, np.arange(100, ncol, 100), axis=0)

        col_begin = 0

        for col, cv_y in enumerate(cv_chunked_col): # pylint: disable=unused-variable
            # To reduce memory, the cost volume is split (along the col axis)
            # into multiple sub-arrays with a step of 100
            cv_chunked_row = np.array_split(cv_y, np.arange(100, nrow, 100), axis=1)
            row_begin = 0
            for row, cv_x in enumerate(cv_chunked_row): # pylint: disable=unused-variable
                disp[col_begin:col_begin + cv_y.shape[0],
                row_begin: row_begin + cv_x.shape[1]] = \
                    cost_volume.coords['disp'].data[np.argmax(cv_x, axis=2)]
                row_begin += cv_x.shape[1]

            col_begin += cv_y.shape[0]

        return disp



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to the Cross Based Cost Aggregation (cbca) method.
"""

from typing import Dict, Union, Tuple, List

import numpy as np
import xarray as xr
from json_checker import Checker, And
from numba import njit

from pandora.filter import AbstractFilter
from pandora.img_tools import shift_right_img
from . import aggregation


@aggregation.AbstractAggregation.register_subclass('cbca')
class CrossBasedCostAggregation(aggregation.AbstractAggregation):
    """
    CrossBasedCostAggregation class, allows to perform the aggregation step
    """
    # Default configuration, do not change these values
    _CBCA_INTENSITY = 30.
    _CBCA_DISTANCE = 5

    def __init__(self, **cfg: dict):
        """
        :param cfg: optional configuration, {'cbca_intensity': value, 'cbca_distance': value}
        :type cfg: dict
        """
        self.cfg = self.check_conf(**cfg)
        self._cbca_intensity = self.cfg['cbca_intensity']
        self._cbca_distance = self.cfg['cbca_distance']

    def check_conf(self, **cfg: Union[str, float, int]) -> Dict[str, Union[str, float, int]]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: aggregation configuration
        :type cfg: dict
        :return cfg: aggregation configuration updated
        :rtype: dict
        """
        # Give the default value if the required element is not in the configuration
        if 'cbca_intensity' not in cfg:
            cfg['cbca_intensity'] = self._CBCA_INTENSITY
        if 'cbca_distance' not in cfg:
            cfg['cbca_distance'] = self._CBCA_DISTANCE

        schema = {
            'aggregation_method': And(str, lambda input: 'cbca'),
            'cbca_intensity': And(float, lambda input: input > 0),
            'cbca_distance': And(int, lambda input: input > 0)
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self):
        """
        Describes the aggregation method
        """
        print('CrossBasedCostAggregation method')

    def cost_volume_aggregation(self, img_left: xr.Dataset, img_right: xr.Dataset, cv: xr.Dataset,
                                **cfg: Union[str, int]) -> None:
        """
        Aggregated the cost volume with Cross-Based Cost Aggregation, using the pipeline define in
        Zhang, K., Lu, J., & Lafruit, G. (2009).
        Cross-based local stereo matching using orthogonal integral images.
        IEEE transactions on circuits and systems for video technology, 19(7), 1073-1079.

        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param cv: cost volume dataset with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :param cfg: images configuration containing the mask convention : valid_pixels, no_data
        :type cfg: dict
        :return: None
        """
        cross_left, cross_right = self.computes_cross_supports(img_left, img_right, cv)

        offset = int(cv.attrs['offset_row_col'])
        # Cost volume has input image size, if offset > 0 do not consider the marge
        if offset > 0:
            cv_data = cv['cost_volume'].data[offset: -offset, offset: -offset]
        else:
            cv_data = cv['cost_volume'].data
        n_col_, n_row_, nb_disp = cv_data.shape

        # Allocate the numpy aggregated cost volume cv = (disp, col, row), for efficient memory management
        agg = np.zeros((nb_disp, n_row_, n_col_), dtype=np.float32)

        # Add invalid costs (i.e = np.nan ) to the output aggregated cost volume (because the step 1 of cbca do not
        # propagate invalid pixels, we need to retrieve them at the end of aggregation )
        # Much faster than :
        # id_nan = np.isnan(cv['cost_volume'].data)
        # compute the aggregation ..
        # cv['cost_volume'].data[id_nan] = np.nan
        agg += np.swapaxes(cv_data, 0, 2)
        agg *= 0

        disparity_range = cv.coords['disp'].data
        range_col = np.arange(0, n_row_)

        for dsp in range(nb_disp):
            i_right = int((disparity_range[dsp] % 1) * cv.attrs['subpixel'])

            # Step 1 : horizontal integral image
            step1 = cbca_step_1(cv_data[:, :, dsp])

            range_col_right = range_col + disparity_range[dsp]
            valid_index = np.where((range_col_right >= 0) & (range_col_right < cross_right[i_right].shape[1]))

            # Step 2 : horizontal matching cost
            step2, sum2 = cbca_step_2(step1, cross_left, cross_right[i_right], range_col[valid_index],
                                      range_col_right[valid_index].astype(int))

            # Step 3 : vertical integral image
            step3 = cbca_step_3(step2)

            # Step 4 : aggregate cost volume
            step4, sum4 = cbca_step_4(step3, sum2, cross_left, cross_right[i_right], range_col[valid_index],
                                      range_col_right[valid_index].astype(int))

            # Added the pixel anchor pixel to the number of support pixels used during the aggregation
            sum4 += 1
            # Add the aggregate cost to the output
            agg[dsp, :, :] += np.swapaxes(step4, 0, 1)
            # Normalize the aggregated cost
            agg[dsp, :, :] /= np.swapaxes(sum4, 0, 1)

        cv_data = np.swapaxes(agg, 0, 2)
        if offset > 0:
            cv['cost_volume'].data[offset: -offset, offset: -offset] = cv_data
        else:
            cv['cost_volume'].data = cv_data
        cv.attrs['aggregation'] = 'cbca'

        # Maximal cost of the cost volume after agregation
        cmax = cv.attrs['cmax'] * ((self._cbca_distance * 2) - 1) ** 2
        cv.attrs['cmax'] = cmax

    def computes_cross_supports(self, img_left: xr.Dataset, img_right: xr.Dataset, cv: xr.Dataset) -> \
            Tuple[np.ndarray, List[np.ndarray]]:
        """
        Prepare images and compute the cross support region of the left and right images.
        A 3x3 median filter is applied to the images before calculating the cross support region.

        :param img_left: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :type img_right: xarray.Dataset
        :param cv: cost volume dataset with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :return: the left and right cross support region
        :rtype: Tuples(left cross support region, List(right cross support region))
        """
        subpix = cv.attrs['subpixel']
        offset = int(cv.attrs['offset_row_col'])

        # shift the right image
        img_right_shift = shift_right_img(img_right, subpix)

        # Median filter on valid pixels
        filter_ = AbstractFilter(**{'filter_method': 'median', 'filter_size': 3})

        # Invalid and no data pixels are masked with np.nan to avoid propagating the values with the median filter
        left_masked = np.copy(img_left['im'].data)
        if 'msk' in img_left.data_vars:
            left_masked[np.where(img_left['msk'].data != img_left.attrs['valid_pixels'])] = np.nan

        left_masked = filter_.median_filter(left_masked)
        # Convert nan to inf to be able to use the comparison operators < and > in cross_support function
        np.nan_to_num(left_masked, copy=False, nan=np.inf)
        # Compute left cross support using numba to reduce running time
        if offset != 0:
            # Cross support to the size of the cost volume
            cross_left = cross_support(left_masked[offset: -offset, offset: -offset], self._cbca_distance,
                                       self._cbca_intensity)
        else:
            cross_left = cross_support(left_masked, self._cbca_distance, self._cbca_intensity)

        # Compute the right cross support. Apply a 3×3 median filter to the input image
        cross_right = []
        for shift, img in enumerate(img_right_shift):
            # Invalid and nodata pixels are masked with np.nan to avoid propagating the values with the median filter
            right_masked = np.copy(img['im'].data)

            # Pixel precision
            if ('msk' in img_right.data_vars) and (shift == 0):
                right_masked[np.where(img_right['msk'].data != img_right.attrs['valid_pixels'])] = np.nan

            # Subpixel precision : computes the shifted right mask
            if ('msk' in img_right.data_vars) and (shift != 0):
                shift_mask = np.zeros(img_right['msk'].data.shape)
                shift_mask[np.where(img_right['msk'].data != img_right.attrs['valid_pixels'])] = np.nan

                # Since the interpolation of the right image is of order 1, the shifted right mask corresponds
                # to an aggregation of two columns of the right mask

                # Create a sliding window of shape 2 using as_strided function : this function create a new a view (by
                # manipulating data pointer)of the shift_mask array with a different shape. The new view pointing to the
                # same memory block as shift_mask so it does not consume any additional memory.
                str_row, str_col = shift_mask.strides  # pylint: disable=unpacking-non-sequence
                shape_windows = (shift_mask.shape[0], shift_mask.shape[1] - 1, 2)
                strides_windows = (str_row, str_col, str_col)
                aggregation_window = np.lib.stride_tricks.as_strided(shift_mask, shape_windows, strides_windows,
                                                                     writeable=False)
                shift_mask = np.sum(aggregation_window, 2)
                right_masked += shift_mask

            #  Apply a 3×3 median filter to the input image
            right_masked = filter_.median_filter(right_masked)
            # Convert nan to inf to be able to use the comparison operators < and > in cross_support function
            np.nan_to_num(right_masked, copy=False, nan=np.inf)
            # Compute right cross support using numba to reduce running time
            if offset != 0:
                # Cross support to the size of the cost volume
                cross_right.append(cross_support(right_masked[offset: -offset, offset: -offset], self._cbca_distance,
                                                 self._cbca_intensity))
            else:
                cross_right.append(cross_support(right_masked, self._cbca_distance, self._cbca_intensity))

        return cross_left, cross_right


@njit('f4[:, :](f4[:, :])', cache=True)
def cbca_step_1(cv: np.ndarray) -> np.ndarray:
    """
    Giving the matching cost for one disparity, build a horizontal integral image storing the cumulative row sum,
    S_h(row, col) = S_h(row-1, col) + cv(row, col)

    :param cv: cost volume for the current disparity
    :type cv: 2D np.array (row, col) dtype = np.float32
    :return: the horizontal integral image, step 1
    :rtype: 2D np.array (row, col + 1) dtype = np.float32
    """
    n_col_, n_row_ = cv.shape
    # Allocate the intermediate cost volume S_h
    # added a column to manage the case in the step 2 : row - left_arm_length -1 = -1
    step1 = np.zeros((n_col_, n_row_ + 1), dtype=np.float32)

    for col in range(n_col_):
        for row in range(n_row_):
            # Do not propagate nan
            if not np.isnan(cv[col, row]):
                step1[col, row] = step1[col, row - 1] + cv[col, row]
            else:
                step1[col, row] = step1[col, row - 1]

    return step1


@njit('(f4[:, :], i2[:, :, :], i2[:, :, :], i8[:], i8[:])', cache=True)
def cbca_step_2(step1: np.ndarray, cross_left: np.ndarray, cross_right: np.ndarray, range_col: np.ndarray,
                range_col_right: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Giving the horizontal integral image, computed the horizontal matching cost for one disparity,
    E_h(row, col) = S_h(row + right_arm_length, col) - S_h(row - left_arm_length -1, col)

    :param step1: horizontal integral image, from the cbca_step1, with an extra column that contains 0
    :type step1: 2D np.array (row, col + 1) dtype = np.float32
    :param cross_left: cross support of the left image
    :type cross_left: 3D np.array (row, col, [left, right, top, bot]) dtype=np.int16
    :param cross_right: cross support of the right image
    :type cross_right: 3D np.array (row, col, [left, right, tpo, bot]) dtype=np.int16
    :param range_col: left column for the current disparity (i.e : np.arrange(nb columns), where the correspondent \
    in the right image is reachable)
    :type range_col: 1D np.array
    :param range_col_right: right column for the current disparity (i.e : np.arrange(nb columns) - disparity, where \
    column - disparity >= 0 and <= nb columns)
    :type range_col_right: 1D np.array
    :return: the horizontal matching cost for the current disparity, and the number of support pixels used for the \
    step 2
    :rtype: tuple (2D np.array (row, col) dtype = np.float32, 2D np.array (row, col) dtype = np.float32)
    """
    n_col_, n_row_ = step1.shape
    # Allocate the intermediate cost volume E_h
    # , remove the extra column from the step 1
    step2 = np.zeros((n_col_, n_row_ - 1), dtype=np.float32)
    sum_step2 = np.zeros((n_col_, n_row_ - 1), dtype=np.float32)

    for col in range(step1.shape[0]):
        for row in range(range_col.shape[0]):
            right = min(cross_left[col, range_col[row], 1], cross_right[col, range_col_right[row], 1])
            left = min(cross_left[col, range_col[row], 0], cross_right[col, range_col_right[row], 0])
            step2[col, range_col[row]] = step1[col, range_col[row] + right] - step1[col, range_col[row] - left - 1]
            sum_step2[col, range_col[row]] += (right + left)

    return step2, sum_step2


@njit('f4[:, :](f4[:, :])', cache=True)
def cbca_step_3(step2: np.ndarray) -> np.ndarray:
    """
    Giving the horizontal matching cost, build a vertical integral image for one disparity,
    S_v = S_v(row, col - 1) + E_h(row, col)

    :param step2: horizontal matching cost, from the cbca_step2
    :type step2: 3D xarray.DataArray (row, col, disp)
    :return: the vertical integral image for the current disparity
    :rtype: 2D np.array (row + 1, col) dtype = np.float32
    """
    n_col_, n_row_ = step2.shape
    # Allocate the intermediate cost volume S_v
    # added a row to manage the case in the step 4 : col - up_arm_length -1 = -1
    step3 = np.zeros((n_col_ + 1, n_row_), dtype=np.float32)
    step3[0, :] = step2[0, :]

    for col in range(1, n_col_):
        for row in range(n_row_):
            step3[col, row] = step3[col - 1, row] + step2[col, row]

    return step3


@njit('(f4[:, :], f4[:, :], i2[:, :, :], i2[:, :, :], i8[:], i8[:])', cache=True)
def cbca_step_4(step3: np.ndarray, sum2: np.ndarray, cross_left: np.ndarray, cross_right: np.ndarray,
                range_col: np.ndarray, range_col_right: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Giving the vertical integral image, build the fully aggregated matching cost for one disparity,
    E = S_v(row, col + bottom_arm_length) - S_v(row, col - top_arm_length - 1)

    :param step3: vertical integral image, from the cbca_step3, with an extra row that contains 0
    :type step3: 2D np.array (row + 1, col) dtype = np.float32
    :param sum2: the number of support pixels used for the step 2
    :type sum2: 2D np.array (row, col) dtype = np.float32
    :param cross_left: cross support of the left image
    :type cross_left: 3D np.array (row, col, [left, right, top, bot]) dtype=np.int16
    :param cross_right: cross support of the right image
    :type cross_right: 3D np.array (row, col, [left, right, tpo, bot]) dtype=np.int16
    :param range_col: left column for the current disparity (i.e : np.arrange(nb columns), where the correspondent \
    in the right image is reachable)
    :type range_col: 1D np.array
    :param range_col_right: right column for the current disparity (i.e : np.arrange(nb columns) - disparity, where \
    column - disparity >= 0 and <= nb columns)
    :type range_col_right: 1D np.array
    :return: the fully aggregated matching cost, and the total number of support pixels used for the aggregation
    :rtype: tuple(2D np.array (row , col) dtype = np.float32, 2D np.array (row , col) dtype = np.float32)
    """
    n_col_, n_row_ = step3.shape
    # Allocate the final cost volume E
    # , remove the extra row from the step 3
    step4 = np.zeros((n_col_ - 1, n_row_), dtype=np.float32)
    sum4 = np.copy(sum2)
    for col in range(step4.shape[0]):
        for row in range(range_col.shape[0]):
            top = min(cross_left[col, range_col[row], 2], cross_right[col, range_col_right[row], 2])
            bot = min(cross_left[col, range_col[row], 3], cross_right[col, range_col_right[row], 3])

            step4[col, range_col[row]] = step3[col + bot, range_col[row]] - step3[col - top - 1, range_col[row]]

            sum4[col, range_col[row]] += (top + bot)
            if top != 0:
                sum4[col, range_col[row]] += np.sum(sum2[col - top:col, range_col[row]])
            if bot != 0:
                sum4[col, range_col[row]] += np.sum(sum2[col + 1:col + bot + 1, range_col[row]])

    return step4, sum4


@njit('i2[:, :, :](f4[:, :], i2, f4)', cache=True)
def cross_support(image: np.ndarray, len_arms: int, intensity: float) -> np.ndarray:
    """
    Compute the cross support for an image: find the 4 arms.
    Enforces a minimum support region of 3×3 if pixels are valid.
    The cross support of invalid pixels (pixels that are np.inf) is 0 for the 4 arms.

    :param image: image
    :type image: 2D np.array (row , col) dtype = np.float32
    :param len_arms: maximal length arms
    :param len_arms: int16
    :param intensity: maximal intensity
    :param intensity: float 32
    :return: a 3D np.array ( row, col, [left, right, top, bot] ), with the four arms lengths computes for each pixel
    :rtype:  3D np.array ( row, col, [left, right, top, bot] ), dtype=np.int16
    """
    n_col_, n_row_ = image.shape
    # By default, all cross supports are 0
    cross = np.zeros((n_col_, n_row_, 4), dtype=np.int16)

    for col in range(n_col_):
        for row in range(n_row_):

            # If the pixel is valid (np.isfinite = True) compute the cross support
            # Else (np.isfinite = False) the pixel is not valid (no data or invalid) and the cross support value is 0
            # for the 4 arms (default value of the variable cross).
            if np.isfinite(image[col, row]):
                left_len = 0
                left = row
                for left in range(row - 1, max(row - len_arms, -1), -1):
                    if abs(image[col, row] - image[col, left]) >= intensity:
                        break
                    left_len += 1
                # enforces a minimum support region of 3×3 if pixels are valid
                cross[col, row, 0] = max(left_len, 1 * (row >= 1) * np.isfinite(image[col, left]))

                right_len = 0
                right = row
                for right in range(row + 1, min(row + len_arms, n_row_)):
                    if abs(image[col, row] - image[col, right]) >= intensity:
                        break
                    right_len += 1
                # enforces a minimum support region of 3×3 if pixels are valid
                cross[col, row, 1] = max(right_len, 1 * (row < n_row_ - 1) * np.isfinite(image[col, right]))

                up_len = 0
                up_col = col
                for up_col in range(col - 1, max(col - len_arms, -1), -1):
                    if abs(image[col, row] - image[up_col, row]) >= intensity:
                        break
                    up_len += 1
                # enforces a minimum support region of 3×3 if pixels are valid
                cross[col, row, 2] = max(up_len, 1 * (col >= 1) * np.isfinite(image[up_col, row]))

                bot_len = 0
                bot = col
                for bot in range(col + 1, min(col + len_arms, n_col_)):
                    if abs(image[col, row] - image[bot, row]) >= intensity:
                        break
                    bot_len += 1
                # enforces a minimum support region of 3×3 if pixels are valid
                cross[col, row, 3] = max(bot_len, 1 * (col < n_col_ - 1) * np.isfinite(image[bot, row]))

    return cross



# pylint:disable=too-many-arguments
# !/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains class associated to the pandora state machine
"""

import logging
from typing import Dict, Union, List
import numpy as np
import xarray as xr
from json_checker import Checker, And

try:
    import graphviz # pylint: disable=unused-import
    from transitions.extensions import GraphMachine as Machine
except ImportError:
    from transitions import Machine
from transitions import MachineError

from pandora import aggregation
from pandora import disparity
from pandora import filter  # pylint: disable=redefined-builtin
from pandora import multiscale
from pandora import optimization
from pandora import refinement
from pandora import matching_cost
from pandora import validation
from .img_tools import prepare_pyramid



# This module contains class associated to the pandora state machine

class PandoraMachine(Machine):  # pylint:disable=too-many-instance-attributes
    """
    PandoraMachine class to create and use a state machine
    """

    _transitions_run = [
        {'trigger': 'matching_cost', 'source': 'begin', 'dest': 'cost_volume', 'after': 'matching_cost_run'},
        {'trigger': 'aggregation', 'source': 'cost_volume', 'dest': 'cost_volume', 'after': 'aggregation_run'},
        {'trigger': 'optimization', 'source': 'cost_volume', 'dest': 'cost_volume', 'after': 'optimization_run'},
        {'trigger': 'disparity', 'source': 'cost_volume', 'dest': 'disp_map', 'after': 'disparity_run'},
        {'trigger': 'filter', 'source': 'disp_map', 'dest': 'disp_map', 'after': 'filter_run'},
        {'trigger': 'refinement', 'source': 'disp_map', 'dest': 'disp_map', 'after': 'refinement_run'},
        {'trigger': 'validation', 'source': 'disp_map', 'dest': 'disp_map', 'after': 'validation_run'},
        # Conditional state change, if it is the last scale the multiscale state will not be triggered
        # This way, after the last scale we can still apply a filter state
        {'trigger': 'multiscale', 'source': 'disp_map', 'conditions': 'is_not_last_scale',
         'dest': 'begin', 'after': 'run_multiscale'},
    ]

    _transitions_check = [
        {'trigger': 'matching_cost', 'source': 'begin', 'dest': 'cost_volume', 'before': 'right_disp_map_check_conf',
         'after': 'matching_cost_check_conf'},
        {'trigger': 'aggregation', 'source': 'cost_volume', 'dest': 'cost_volume', 'after': 'aggregation_check_conf'},
        {'trigger': 'optimization', 'source': 'cost_volume', 'dest': 'cost_volume', 'after': 'optimization_check_conf'},
        {'trigger': 'disparity', 'source': 'cost_volume', 'dest': 'disp_map', 'after': 'disparity_check_conf'},
        {'trigger': 'filter', 'source': 'disp_map', 'dest': 'disp_map', 'after': 'filter_check_conf'},
        {'trigger': 'refinement', 'source': 'disp_map', 'dest': 'disp_map', 'after': 'refinement_check_conf'},
        {'trigger': 'validation', 'source': 'disp_map', 'dest': 'disp_map', 'after': 'validation_check_conf'},
        # For the check conf we define the destination of multiscale state as disp_map instead of begin
        # given the conditional change of state
        {'trigger': 'multiscale', 'source': 'disp_map', 'dest': 'disp_map', 'after': 'multiscale_check_conf'}
    ]

    def __init__(self, img_left_pyramid: List[xr.Dataset] = None, img_right_pyramid: List[xr.Dataset] = None,
                 disp_min: Union[np.array, int] = None, disp_max: Union[np.array, int] = None,
                 right_disp_min: Union[np.array, None] = None, right_disp_max: Union[np.array, None] = None
                 ) -> None:

        """
        Initialize Pandora Machine

        :param img_left_pyramid: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :type img_left_pyramid: xarray.Dataset
        :param img_right_pyramid: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :type img_right_pyramid: xarray.Dataset
        :param disp_min: minimal disparity
        :type disp_min: int or np.array
        :param disp_max: maximal disparity
        :type disp_max: int or np.array
        :param right_disp_min: minimal disparity of the right image
        :type right_disp_min: None or np.array
        :param right_disp_max: maximal disparity of the right image
        :type right_disp_max: None or np.array
        :return: None
        """
        # Left image scale pyramid
        self.img_left_pyramid = img_left_pyramid
        # Right image scale pyramid
        self.img_right_pyramid = img_right_pyramid
        # Left image
        self.left_img = None
        # Right image
        self.right_img = None
        # Minimum disparity
        self.disp_min = disp_min
        # Maximum disparity
        self.disp_max = disp_max
        # Maximum disparity for the right image
        self.right_disp_min = right_disp_min
        # Minimum disparity for the right image
        self.right_disp_max = right_disp_max
        # User minimum disparity
        self.dmin_user = None
        # User maximum disparity
        self.dmax_user = None
        # User minimum disparity right
        self.dmin_user_right = None
        # User maximum disparity right
        self.dmax_user_right = None

        # Scale factor
        self.scale_factor = None
        # Number of scales
        self.num_scales = 1
        # Current scale
        self.current_scale = None

        # left cost volume
        self.left_cv = None
        # right cost volume
        self.right_cv = None
        # left disparity map
        self.left_disparity = None
        # right disparity map
        self.right_disparity = None

        # Pandora's pipeline configuration
        self.pipeline_cfg = {'pipeline': {}}
        # Right disparity map computation information: Can be "none" or "accurate"
        # Useful during the running steps to choose if we must compute left and right objects.
        self.right_disp_map = 'none'
        # Define avalaible states
        states_ = ['begin', 'cost_volume', 'disp_map']

        # Initialize a machine without any transition
        Machine.__init__(self, states=states_, initial='begin', transitions=None, auto_transitions=False)

        logging.getLogger('transitions').setLevel(logging.WARNING)

    def matching_cost_run(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
        Matching cost computation
        :param cfg: pipeline configuration
        :type  cfg: dict
        :param input_step: step to trigger
        :type input_step: str
        :return: None
        """

        logging.info('Matching cost computation...')
        matching_cost_ = matching_cost.AbstractMatchingCost(**cfg[input_step])

        # Update min and max disparity according to the current scale
        self.disp_min = self.disp_min * self.scale_factor
        self.disp_max = self.disp_max * self.scale_factor
        # Obtain absolute min and max disparities
        dmin_min, dmax_max = matching_cost_.dmin_dmax(self.disp_min, self.disp_max)

        # Compute cost volume and mask it
        if self.right_disp_map != 'accurate':
            self.left_cv = matching_cost_.compute_cost_volume(self.left_img, self.right_img, dmin_min, dmax_max)
            matching_cost_.cv_masked(self.left_img, self.right_img, self.left_cv, self.disp_min, self.disp_max)

        else:
            self.left_cv = matching_cost_.compute_cost_volume(self.left_img, self.right_img, dmin_min, dmax_max)
            matching_cost_.cv_masked(self.left_img, self.right_img, self.left_cv, self.disp_min, self.disp_max)

            # Update min and max disparity according to the current scale
            self.right_disp_min = self.right_disp_min * self.scale_factor
            self.right_disp_max = self.right_disp_max * self.scale_factor
            # Obtain absolute min and max right disparities
            dmin_min_right, dmax_max_right = matching_cost_.dmin_dmax(self.right_disp_min, self.right_disp_max)
            # Compute right cost volume and mask it
            self.right_cv = matching_cost_.compute_cost_volume(self.right_img, self.left_img, dmin_min_right,
                                                               dmax_max_right)
            matching_cost_.cv_masked(self.right_img, self.left_img, self.right_cv, self.right_disp_min,
                              self.right_disp_max)

    def aggregation_run(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
         Cost (support) aggregation
        :param cfg: pipeline configuration
        :type  cfg: dict
        :param input_step: step to trigger
        :type input_step: str
        :return: None
        """
        aggregation_ = aggregation.AbstractAggregation(**cfg[input_step])
        if self.right_disp_map != 'accurate':
            aggregation_.cost_volume_aggregation(self.left_img, self.right_img, self.left_cv)
        else:
            aggregation_.cost_volume_aggregation(self.left_img, self.right_img, self.left_cv)
            aggregation_.cost_volume_aggregation(self.right_img, self.left_img, self.right_cv)

    def optimization_run(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
         Cost optimization
        :param cfg: pipeline configuration
        :type  cfg: dict
        :param input_step: step to trigger
        :type input_step: str
        :return: None
        """
        optimization_ = optimization.AbstractOptimization(**cfg[input_step])
        logging.info('Cost optimization...')
        if self.right_disp_map != 'accurate':
            self.left_cv = optimization_.optimize_cv(self.left_cv, self.left_img, self.right_img)
        else:
            self.left_cv = optimization_.optimize_cv(self.left_cv, self.left_img, self.right_img)
            self.right_cv = optimization_.optimize_cv(self.right_cv, self.right_img, self.left_img)

    def disparity_run(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
        Disparity computation and validity mask
        :param cfg: pipeline configuration
        :type  cfg: dict
        :param input_step: step to trigger
        :type input_step: str
        :return: None
        """
        logging.info('Disparity computation...')
        disparity_ = disparity.AbstractDisparity(**cfg[input_step])
        if self.right_disp_map != 'accurate':
            self.left_disparity = disparity_.to_disp(self.left_cv, self.left_img, self.right_img)
            disparity_.validity_mask(self.left_disparity, self.left_img, self.right_img,
                                     self.left_cv)
        elif self.right_disp_map == 'accurate':
            self.left_disparity = disparity_.to_disp(self.left_cv, self.left_img, self.right_img)
            disparity_.validity_mask(self.left_disparity, self.left_img, self.right_img,
                                     self.left_cv)
            self.right_disparity = disparity_.to_disp(self.right_cv, self.right_img, self.left_img)
            disparity_.validity_mask(self.right_disparity, self.right_img, self.left_img,
                                     self.right_cv)

    def filter_run(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
        Disparity filter
        :param cfg: pipeline configuration
        :type  cfg: dict
        :param input_step: step to trigger
        :type input_step: str
        :return: None
        """
        logging.info('Disparity filtering...')
        filter_ = filter.AbstractFilter(**cfg[input_step])
        if self.right_disp_map != 'accurate':
            filter_.filter_disparity(self.left_disparity)
        else:
            filter_.filter_disparity(self.left_disparity)
            filter_.filter_disparity(self.right_disparity)

    def refinement_run(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
         Subpixel disparity refinement
        :param cfg: pipeline configuration
        :type  cfg: dict
        :param input_step: step to trigger
        :type input_step: str
        :return: None
        """
        refinement_ = refinement.AbstractRefinement(**cfg[input_step])
        logging.info('Subpixel refinement...')
        if self.right_disp_map != 'accurate':
            refinement_.subpixel_refinement(self.left_cv, self.left_disparity)
        else:
            refinement_.subpixel_refinement(self.left_cv, self.left_disparity)
            refinement_.subpixel_refinement(self.right_cv, self.right_disparity)

    def validation_run(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
         Validation of disparity map
        :param cfg: pipeline configuration
        :type  cfg: dict
        :param input_step: step to trigger
        :type input_step: str
        :return: None
        """
        validation_ = validation.AbstractValidation(**cfg[input_step])

        logging.info('Validation...')

        if self.right_disp_map != 'accurate':
            self.left_disparity = validation_.disparity_checking(self.left_disparity, self.right_disparity)

        else:
            self.left_disparity = validation_.disparity_checking(self.left_disparity, self.right_disparity)
            self.right_disparity = validation_.disparity_checking(self.right_disparity, self.left_disparity)
            # Interpolated mismatch and occlusions
            if 'interpolated_disparity' in cfg:
                interpolate_ = validation.AbstractInterpolation(**cfg[input_step])
                interpolate_.interpolated_disparity(self.left_disparity)
                interpolate_.interpolated_disparity(self.right_disparity)

    def run_multiscale(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
            Compute the disparity range for the next scale
            :param cfg: pipeline configuration
            :type  cfg: dict
            :param input_step: step to trigger
            :type input_step: str
            :return: None
            """

        logging.info('Disparity range computation...')

        multiscale_ = multiscale.AbstractMultiscale(**cfg[input_step])

        # Update min and max user disparity according to the current scale
        self.dmin_user = self.dmin_user * self.scale_factor
        self.dmax_user = self.dmax_user * self.scale_factor

        # Compute disparity range for the next scale level
        if self.right_disp_map != 'accurate':
            self.disp_min, self.disp_max = multiscale_.disparity_range(self.left_disparity,
                                                                        self.dmin_user, self.dmax_user)
        else:
            self.disp_min, self.disp_max = multiscale_.disparity_range(self.left_disparity,
                                                                        self.dmin_user, self.dmax_user)
            # Update min and max user disparity according to the current scale
            self.dmin_user_right = self.dmin_user_right * self.scale_factor
            self.dmax_user_right = self.dmax_user_right * self.scale_factor
            self.right_disp_min, self.right_disp_max = multiscale_.disparity_range(self.right_disparity,
                                                                                    self.dmin_user_right,
                                                                                    self.dmax_user_right)

        # Get the next scale's images
        self.left_img = self.img_left_pyramid.pop(0)
        self.right_img = self.img_right_pyramid.pop(0)

        # Update the current scale for the next state
        self.current_scale = self.current_scale - 1

    def run_prepare(self, cfg: Dict[str, dict], left_img: xr.Dataset, right_img: xr.Dataset,
                    disp_min: Union[np.array, int], disp_max: Union[np.array, int], scale_factor: int,
                    num_scales: int, right_disp_min: Union[None, np.array] = None,
                    right_disp_max: Union[None, np.array] = None) -> None:
        """
        Prepare the machine before running

        :param cfg:  configuration
        :type cfg: dict
        :param left_img: left Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :type left_img: xarray.Dataset
        :param right_img: right Dataset image containing :

                - im : 2D (row, col) xarray.DataArray
                - msk (optional): 2D (row, col) xarray.DataArray
        :type right_img: xarray.Dataset
        :param disp_min: minimal disparity
        :type disp_min: int or np.array
        :param disp_max: maximal disparity
        :type disp_max: int or np.array
        :param scale_factor: scale factor for multiscale
        :type scale_factor: int
        :param num_scales: scales number for multiscale
        :type num_scales: int
        :param disp_min_right: minimal disparity of the right image
        :type disp_min_right: np.array or None
        :param disp_max_right: maximal disparity of the right image
        :type disp_max_right: np.array or None
        :return: None
        """
        # Total number of scales
        self.num_scales = num_scales
        # Scale factor
        self.scale_factor = scale_factor


        if self.num_scales > 1:
            # If multiscale processing, create pyramid and select first scale's images
            self.img_left_pyramid, self.img_right_pyramid = prepare_pyramid(left_img, right_img, self.num_scales,
                                                                            scale_factor)
            self.left_img = self.img_left_pyramid.pop(0)
            self.right_img = self.img_right_pyramid.pop(0)
            # Initialize current scale
            self.current_scale = num_scales - 1
            # If multiscale, disparities can only be int.
            # Downscale disparities since the pyramid is processed from coarse to original size
            self.disp_min = int(disp_min / (scale_factor ** self.num_scales))
            self.disp_max = int(disp_max / (scale_factor ** self.num_scales))
            # User disparity
            self.dmin_user = self.disp_min
            self.dmax_user = self.disp_max
            # If multiscale disparities can only be int, and right disparity can only be np.array, so right disparity
            # can not be defined in the input conf
            # Right disparities
            self.right_disp_min = -self.disp_max
            self.right_disp_max = -self.disp_min
            # Right user disparity
            self.dmin_user_right = self.right_disp_min
            self.dmax_user_right = self.right_disp_max
        else:
            # If no multiscale processing, select the original images
            self.left_img = left_img
            self.right_img = right_img
            # If no multiscale processing, current scale is zero
            self.current_scale = 0
            # Disparities
            self.disp_min = disp_min
            self.disp_max = disp_max
            # Right disparities
            if right_disp_max is not None and right_disp_min is not None:
                # If the right disparity was defined in the input conf
                self.right_disp_min = right_disp_min
                self.right_disp_max = right_disp_max
            else:
                self.right_disp_min = -self.disp_max
                self.right_disp_max = -self.disp_min


        # Initiate output disparity datasets
        self.left_disparity = xr.Dataset()
        self.right_disparity = xr.Dataset()
        # To determine whether the right disparity map has to be computed
        self.right_disp_map = cfg['right_disp_map']['method']
        # Add transitions
        self.add_transitions(self._transitions_run)

    def run(self, input_step: str, cfg: Dict[str, dict]) -> None:
        """
        Run pandora step by triggering the corresponding machine transition

        :param input_step: step to trigger
        :type input_step: str
        :param cfg: pipeline configuration
        :type  cfg: dict
        :return: None
        """

        try:
            # There may be steps that are repeated several times, for example:
            #     'filter': {
            #       'filter_method': 'median'
            #     },
            #     'filter.1': {
            #       'filter_method': 'bilateral'
            #     }
            # But there's only a filter transition. Therefore, in the case of filter.1 we have to call the
            # filter
            # trigger and give the configuration of filter.1
            if len(input_step.split('.')) != 1:
                self.trigger(input_step.split('.')[0], cfg, input_step)
            else:
                self.trigger(input_step, cfg, input_step)
        except (MachineError, KeyError):
            print('\n A problem occurs during Pandora running ' + input_step +
                  '. Be sure of your sequencement step  \n')
            raise

    def run_exit(self) -> None:
        """
        Clear transitions and return to state begin

        :return: None
        """
        self.remove_transitions(self._transitions_run)
        self.set_state('begin')

    def right_disp_map_check_conf(self, cfg: Dict[str, dict],
                                  input_step: str) -> None:  # pylint:disable=unused-argument
        """
        Check the right_disp_map configuration

        :param cfg: configuration
        :type cfg: dict
        :param input_step: current step
        :type input_step: string
        :return: None
        """

        schema = {
            'method': And(str, lambda input: 'none' or 'accurate')
        }

        checker = Checker(schema)
        checker.validate(cfg['right_disp_map'])

        # Store the righ disparity map configuration
        self.right_disp_map = cfg['right_disp_map']['method']

    def matching_cost_check_conf(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
        Check the matching cost configuration

        :param cfg: configuration
        :type cfg: dict
        :param input_step: current step
        :type input_step: string
        :return: None
        """

        matching_cost_ = matching_cost.AbstractMatchingCost(**cfg[input_step])
        self.pipeline_cfg['pipeline'][input_step] = matching_cost_.cfg

    def disparity_check_conf(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
        Check the disparity computation configuration

        :param cfg: configuration
        :type cfg: dict
        :param input_step: current step
        :type input_step: string
        :return: None
        """
        disparity_ = disparity.AbstractDisparity(**cfg[input_step])
        self.pipeline_cfg['pipeline'][input_step] = disparity_.cfg

    def filter_check_conf(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
        Check the filter configuration

        :param cfg: configuration
        :type cfg: dict
        :param input_step: current step
        :type input_step: string
        :return: None
        """
        filter_ = filter.AbstractFilter(**cfg[input_step])
        self.pipeline_cfg['pipeline'][input_step] = filter_.cfg

    def refinement_check_conf(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
        Check the refinement configuration

        :param cfg: configuration
        :type cfg: dict
        :param input_step: current step
        :type input_step: string
        :return: None
        """
        refinement_ = refinement.AbstractRefinement(**cfg[input_step])
        self.pipeline_cfg['pipeline'][input_step] = refinement_.cfg

    def aggregation_check_conf(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
        Check the aggregation configuration

        :param cfg: configuration
        :type cfg: dict
        :param input_step: current step
        :type input_step: string
        :return: None
        """
        aggregation_ = aggregation.AbstractAggregation(**cfg[input_step])
        self.pipeline_cfg['pipeline'][input_step] = aggregation_.cfg

    def optimization_check_conf(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
        Check the optimization configuration

        :param cfg: configuration
        :type cfg: dict
        :param input_step: current step
        :type input_step: string
        :return: None
        """
        optimization_ = optimization.AbstractOptimization(**cfg[input_step])
        self.pipeline_cfg['pipeline'][input_step] = optimization_.cfg

    def validation_check_conf(self, cfg: Dict[str, dict], input_step: str) -> None:
        """
        Check the validation configuration

        :param cfg: configuration
        :type cfg: dict
        :param input_step: current step
        :type input_step: string
        :return: None
        """

        validation_ = validation.AbstractValidation(**cfg[input_step])
        self.pipeline_cfg['pipeline'][input_step] = validation_.cfg
        if 'interpolated_disparity' in validation_.cfg:
            interpolate_ = validation.AbstractInterpolation(**cfg[input_step])  # pylint: disable=unused-variable

        if validation_.cfg['validation_method'] == 'cross_checking' and self.right_disp_map != 'accurate':
            raise MachineError('Can t trigger event cross-checking validation  if right_disp_map method equal to '
                               + self.right_disp_map)

    def multiscale_check_conf(self, cfg: Dict[str, dict], input_step: str):
        """
        Check the disparity computation configuration

        :param cfg: disparity computation configuration
        :type cfg: dict
        :param input_step: current step
        :type input_step: string
        :return:
        """
        multiscale_ = multiscale.AbstractMultiscale(**cfg[input_step])
        self.pipeline_cfg['pipeline'][input_step] = multiscale_.cfg

    def check_conf(self, cfg: Dict[str, dict]):
        """
        Check configuration and transitions

        :param cfg: pipeline configuration
        :type  cfg: dict
        :return:
        """

        # Add transitions to the empty machine.
        self.add_transitions(self._transitions_check)

        # Warning: first element of cfg dictionary is not a transition. It contains information about the way to
        # compute right disparity map.
        for input_step in list(cfg)[1:]:
            try:
                # There may be steps that are repeated several times, for example:
                #     'filter': {
                #       'filter_method': 'median'
                #     },
                #     'filter.1': {
                #       'filter_method': 'bilateral'
                #     }
                # But there's only a filter transition. Therefore, in the case of filter.1 we have to call the
                # filter
                # trigger and give the configuration of filter.1

                if len(input_step.split('.')) != 1:
                    self.trigger(input_step.split('.')[0], cfg, input_step)
                else:
                    self.trigger(input_step, cfg, input_step)

            except (MachineError, KeyError):
                print('\n Problem during Pandora checking configuration steps sequencing. '
                      'Check your configuration file. \n')
                raise

        # Remove transitions
        self.remove_transitions(self._transitions_check)

        # Coming back to the initial state
        self.set_state('begin')

    def remove_transitions(self, transition_list: Dict[str, dict]) -> None:
        """
        Delete all transitions defined in the input list

        :param transition_list: list of transitions
        :type transition_list: dict
        :return: None
        """
        # Transition is removed using trigger name. But one trigger name can be used by multiple transitions
        # In this case, the 'remove_transition' function removes all transitions using this trigger name
        # deleted_triggers list is used to avoid multiple call of 'remove_transition'' with the same trigger name.
        deleted_triggers = []
        for trans in transition_list:
            if trans['trigger'] not in deleted_triggers:
                self.remove_transition(trans['trigger'])
                deleted_triggers.append(trans['trigger'])

    def is_not_last_scale(self, input_step: str, cfg: Dict[str, dict]):  # pylint:disable=unused-argument
        """
        Check if the current scale is the last scale
        :param cfg: configuration
        :type cfg: dict
        :param input_step: current step
        :type input_step: string
        :return: boolean
        """

        if self.current_scale == 0:
            return False
        return True



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to the quadratic method used in the refinement step.
"""

from typing import Dict, Tuple

import numpy as np
from json_checker import Checker, And
from numba import njit

import pandora.constants as cst
from . import refinement


@refinement.AbstractRefinement.register_subclass('quadratic')
class Quadratic(refinement.AbstractRefinement):
    """
    Quadratic class allows to perform the subpixel cost refinement step
    """

    def __init__(self, **cfg: str) -> None:
        """
        :param cfg: optional configuration, {}
        :type cfg: dict
        :return: None
        """
        self.cfg = self.check_conf(**cfg)
        self._refinement_method_name = str(self.cfg['refinement_method'])

    @staticmethod
    def check_conf(**cfg: str) -> Dict[str, str]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: refinement configuration
        :type cfg: dict
        :return cfg: refinement configuration updated
        :rtype: dict
        """
        schema = {
            'refinement_method': And(str, lambda input: 'quadratic')
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self) -> None:
        """
        Describes the subpixel refinement method
        :return: None
        """
        print('Quadratic refinement method')

    @staticmethod
    @njit(cache=True)
    def refinement_method(cost: np.ndarray, disp: float, measure: str) -> Tuple[float, float, int]:
        """
        Return the subpixel disparity and cost, by fitting a quadratic curve

        :param cost: cost of the values disp - 1, disp, disp + 1
        :type cost: 1D numpy array : [cost[disp -1], cost[disp], cost[disp + 1]]
        :param disp: the disparity
        :type disp: float
        :param measure: the type of measure used to create the cost volume
        :param measure: string = min | max
        :return: the refined disparity (disp + sub_disp), the refined cost and the state of the pixel ( Information: \
        calculations stopped at the pixel step, sub-pixel interpolation did not succeed )
        :rtype: float, float, int
        """

        if (np.isnan(cost[0])) or (np.isnan(cost[2])):
            # Bit 3 = 1: Information: calculations stopped at the pixel step, sub-pixel interpolation did not succeed
            return disp, cost[1], cst.PANDORA_MSK_PIXEL_STOPPED_INTERPOLATION

        # Solve the system: col = alpha * row ** 2 + beta * row + gamma
        alpha = (cost[0] - 2 * cost[1] + cost[2]) / 2
        beta = (cost[2] - cost[0]) / 2
        gamma = cost[1]

        # If the costs are close, the result of -b / 2a (minimum) is bounded between [-1, 1]
        # sub_disp is row
        sub_disp = min(1.0, max(-1.0, -beta / (2 * alpha)))

        # sub_cost is col
        sub_cost = (alpha * sub_disp ** 2) + (beta * sub_disp) + gamma

        return disp + sub_disp, sub_cost, 0



# pylint: disable=invalid-name
#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains the general function to run Pandora pipeline.
"""

import argparse

import pandora


def get_parser():
    """
    ArgumentParser for Pandora

    :return parser
    """
    parser = argparse.ArgumentParser(description='Pandora stereo matching')
    parser.add_argument('config', help='Path to a json file containing the input files paths and algorithm parameters')
    parser.add_argument('output_dir', help='Path to the output directory')
    parser.add_argument('-v', '--verbose', help='Increase output verbosity', action='store_true')
    return parser


def main():
    """
    Call Pandora's main
    """
    # Get parser
    parser = get_parser()
    args = parser.parse_args()

    # Run the Pandora pipeline
    pandora.main(args.config, args.output_dir, args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains the required libraries and softwares allowing to execute the software,
and setup elements to configure and identify the software.
"""

from codecs import open as copen
from setuptools import setup, find_packages

CMDCLASS = {}

try:
    from sphinx.setup_command import BuildDoc

    CMDCLASS['build_sphinx'] = BuildDoc
except ImportError:
    print('WARNING: sphinx not available. Doc cannot be built')

REQUIREMENTS = ['numpy',
                'xarray>=0.13.*',
                'scipy',
                'rasterio',
                'json-checker',
                'numba>=0.47.*',
                'opencv-python-headless',
                'transitions',
                'scikit-image']


REQUIREMENTS_EXTRA = {'dev': ['sphinx',
                              'sphinx_rtd_theme',
                              'sphinx_autoapi',
                              'nose2',
                              'pylint',
                              'pre-commit'],
                      'sgm': ['pandora_plugin_libsgm==0.6.*'],
                      'docs': ['sphinx',
                               'sphinx_rtd_theme',
                               'sphinx_autoapi']
                      }

def readme():
    with copen('README.md', 'r', 'utf-8') as fstream:
        return fstream.read()


setup(name='pandora',
      version='0.5.0',
      description='Pandora is a stereo matching framework that helps emulate state of the art algorithms',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://github.com/CNES/Pandora',
      author='CNES',
      author_email='myriam.cournet@cnes.fr',
      license='Apache License 2.0',
      entry_points={
          'console_scripts': ['pandora = pandora.Pandora:main']
      },
      python_requires='>=3.6',
      install_requires=REQUIREMENTS,
      extras_require=REQUIREMENTS_EXTRA,
      packages=find_packages(),
      cmdclass=CMDCLASS,
      command_options={
          'build_sphinx': {
              'build_dir': ('setup.py', 'doc/build/'),
              'source_dir': ('setup.py', 'doc/sources/')}})



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to the bilateral filter used to filter the disparity map.
"""

from typing import Dict, Union

import cv2
import xarray as xr
from json_checker import Checker, And

from . import filter # pylint: disable=redefined-builtin


@filter.AbstractFilter.register_subclass('bilateral')
class BilateralFilter(filter.AbstractFilter):
    """
    BilateralFilter class allows to perform the filtering step
    """
    # Default configuration, do not change these values
    _SIGMA_COLOR = 2.
    _SIGMA_SPACE = 6.

    def __init__(self, **cfg: Union[str, float]):
        """
        :param cfg: optional configuration, {'sigmaColor' : value, 'sigmaSpace' : value}
        :type cfg: dict
        """
        self.cfg = self.check_conf(**cfg)
        self._sigma_color = self.cfg['sigma_color']
        self._sigma_space = self.cfg['sigma_space']

    def check_conf(self, **cfg: Union[str, float]) -> Dict[str, Union[str, float]]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: filter configuration
        :type cfg: dict
        :return cfg: filter configuration updated
        :rtype: dict
        """
        # Give the default value if the required element is not in the configuration
        if 'sigma_color' not in cfg:
            cfg['sigma_color'] = self._SIGMA_COLOR
        if 'sigma_space' not in cfg:
            cfg['sigma_space'] = self._SIGMA_SPACE

        schema = {
            'filter_method': And(str, lambda input: 'bilateral'),
            'sigma_color': And(float, lambda input: input > 0),
            'sigma_space': And(float, lambda input: input > 0)
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self):
        """
        Describes the filtering method
        """
        print('Bilateral filter description')

    def filter_disparity(self, disp: xr.Dataset, img_left: xr.Dataset = None, img_right: xr.Dataset = None,
                         cv: xr.Dataset = None) -> None:
        """
        Apply bilateral filter using openCV.
        Filter size is computed from sigmaSpace

        :param disp: the disparity map dataset  with the variables :

                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
                - validity_mask 2D xarray.DataArray (row, col)
        :type disp: xarray.Dataset
        :param img_left: left Dataset image
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image
        :type img_right: xarray.Dataset
        :param cv: cost volume dataset
        :type cv: xarray.Dataset
        :return: None
        """
        disp['disparity_map'].data = cv2.bilateralFilter(disp['disparity_map'].data, d=0, sigmaColor=self._sigma_color,
                                                         sigmaSpace=self._sigma_space)
        disp.attrs['filter'] = 'bilateral'



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains classes and functions associated to the disparity map filtering.
"""

import logging
import sys
from abc import ABCMeta, abstractmethod

import xarray as xr


class AbstractFilter():
    """
    Abstract Filter class
    """
    __metaclass__ = ABCMeta

    filter_methods_avail = {}

    def __new__(cls, **cfg: dict) -> object:
        """
        Return the plugin associated with the filter_method given in the configuration

        :param cfg: the configuration {'filter_method': value}
        :type cfg: dictionary
        """
        if cls is AbstractFilter:
            if isinstance(cfg['filter_method'], str):
                try:
                    return super(AbstractFilter, cls).__new__(cls.filter_methods_avail[cfg['filter_method']])
                except KeyError:
                    logging.error('No filter method named % supported', cfg['filter_method'])
                    sys.exit(1)
            else:
                if isinstance(cfg['filter_method'], unicode): # pylint: disable=undefined-variable
                    # creating a plugin from registered short name given as unicode (py2 & 3 compatibility)
                    try:
                        return super(AbstractFilter, cls).__new__(
                            cls.filter_methods_avail[cfg['filter_method'].encode('utf-8')])
                    except KeyError:
                        logging.error('No filter method named % supported', cfg['filter_method'])
                        sys.exit(1)
        else:
            return super(AbstractFilter, cls).__new__(cls)

    @classmethod
    def register_subclass(cls, short_name: str):
        """
        Allows to register the subclass with its short name

        :param short_name: the subclass to be registered
        :type short_name: string
        """

        def decorator(subclass):
            """
            Registers the subclass in the available methods

            :param subclass: the subclass to be registered
            :type subclass: object
            """
            cls.filter_methods_avail[short_name] = subclass
            return subclass

        return decorator

    @abstractmethod
    def desc(self):
        """
        Describes the filtering method

        """
        print('Filtering method description')

    @abstractmethod
    def filter_disparity(self, disp: xr.Dataset, img_left: xr.Dataset = None, img_right: xr.Dataset = None,
                         cv: xr.Dataset = None) -> None:
        """
        Post processing the disparity map by applying a filter on valid pixels

        :param disp: the disparity map dataset with the variables :

                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
                - validity_mask 2D xarray.DataArray (row, col)
        :type disp: xarray.Dataset
        :param img_left: left Dataset image
        :type img_left: xarray.Dataset
        :param img_right: right Dataset image
        :type img_right: xarray.Dataset
        :param cv: cost volume dataset
        :type cv: xarray.Dataset
        :return: None
        """



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions to run Pandora pipeline.
"""

import logging
import logging.config
from typing import Dict, Tuple, Union

import xarray as xr
import numpy as np
from pkg_resources import iter_entry_points

from . import common
from .img_tools import read_img, read_disp
from .check_json import check_conf, read_config_file, read_multiscale_params
from .state_machine import PandoraMachine


# pylint: disable=too-many-arguments
def run(pandora_machine: PandoraMachine, img_left: xr.Dataset, img_right: xr.Dataset, disp_min: Union[np.array, int],
        disp_max: Union[np.array, int], cfg: Dict[str, dict], disp_min_right: Union[None, np.array] = None,
        disp_max_right: Union[None, np.array] = None) -> Tuple[xr.Dataset, xr.Dataset]:
    """
    Run the pandora pipeline

    :param pandora_machine: instance of PandoraMachine
    :type pandora_machine: PandoraMachine
    :param img_left: left Dataset image containing :

            - im : 2D (row, col) xarray.DataArray
            - msk (optional): 2D (row, col) xarray.DataArray
    :type img_left: xarray.Dataset
    :param img_right: right Dataset image containing :

            - im : 2D (row, col) xarray.DataArray
            - msk (optional): 2D (row, col) xarray.DataArray
    :type img_right: xarray.Dataset
    :param disp_min: minimal disparity
    :type disp_min: int or np.array
    :param disp_max: maximal disparity
    :type disp_max: int or np.array
    :param cfg: pipeline configuration
    :type cfg: Dict[str, dict]
    :param disp_min_right: minimal disparity of the right image
    :type disp_min_right: np.array or None
    :param disp_max_right: maximal disparity of the right image
    :type disp_max_right: np.array or None
    :return: Two xarray.Dataset :


            - left : the left dataset, which contains the variables :
                - disparity_map : the disparity map in the geometry of the left image 2D DataArray (row, col)
                - confidence_measure : the confidence measure in the geometry of the left image 3D DataArray \
                    (row, col, indicator)
                - validity_mask : the validity mask in the geometry of the left image 2D DataArray (row, col)

            - right : the right dataset. If there is no validation step, the right Dataset will be empty.If a \
            validation step is configured, the dataset will contain the variables :
                - disparity_map : the disparity map in the geometry of the right image 2D DataArray (row, col)
                - confidence_measure : the confidence measure in the geometry of the left image 3D DataArray \
                    (row, col, indicator)
                - validity_mask : the validity mask in the geometry of the left image 2D DataArray (row, col)

    :rtype: tuple (xarray.Dataset, xarray.Dataset)
    """

    # Retrieve the multiscale parameter from the conf. If no multiscale was defined, num_scales=0 and scale_factor=1
    num_scales, scale_factor = read_multiscale_params(cfg)

    # Prepare the machine before running, including the image pyramid creation
    pandora_machine.run_prepare(cfg, img_left, img_right, disp_min, disp_max, scale_factor, num_scales, disp_min_right,
                                disp_max_right)

    # For each scale we run the whole pipeline, the scales will be executed from coarse to fine,
    # and the machine will store the necessary parameters for the following scale
    for scale in range(pandora_machine.num_scales): #pylint:disable=unused-variable
        # Trigger the machine step by step
        for elem in list(cfg)[1:]:
            pandora_machine.run(elem, cfg)
            # If the machine gets to the begin state, pass to next scale
            if pandora_machine.state == 'begin':
                break

    # Stop the machine which returns to its initial state
    pandora_machine.run_exit()

    return pandora_machine.left_disparity, pandora_machine.right_disparity


def setup_logging(verbose: bool) -> None:
    """
    Setup the logging configuration

    :param verbose: verbose mode
    :type verbose: bool
    :return: None
    """
    if verbose:
        logging.basicConfig(format='[%(asctime)s][%(levelname)s] %(message)s', level=logging.INFO)
    else:
        logging.basicConfig(format='[%(asctime)s][%(levelname)s] %(message)s', level=logging.ERROR)

def import_plugin() -> None:
    """
    Load all the registered entry points
    :return: None
    """
    for entry_point in iter_entry_points(group='pandora.plugin'):
        entry_point.load()


def main(cfg_path: str, output: str, verbose: bool) -> None:
    """
    Check config file and run pandora framework accordingly

    :param cfg_path: path to the json configuration file
    :type cfg_path: string
    :param output: Path to output directory
    :type output: string
    :param verbose: verbose mode
    :type verbose: bool
    :return: None
    """

    # Read the user configuration file
    user_cfg = read_config_file(cfg_path)

    # Import pandora plugins
    import_plugin()

    # Instantiate pandora state machine
    pandora_machine = PandoraMachine()
    # check the configuration
    cfg = check_conf(user_cfg, pandora_machine)

    # setup the logging configuration
    setup_logging(verbose)

    # Read images and masks
    img_left = read_img(cfg['input']['img_left'], no_data=cfg['input']['nodata_left'],mask=cfg['input']['left_mask'],
                        classif=cfg['input']['left_classif'], segm=cfg['input']['left_segm'])
    img_right = read_img(cfg['input']['img_right'], no_data=cfg['input']['nodata_right'],
                         mask=cfg['input']['right_mask'], classif=cfg['input']['right_classif'],
                         segm=cfg['input']['right_segm'])

    # Read range of disparities
    disp_min = read_disp(cfg['input']['disp_min'])
    disp_max = read_disp(cfg['input']['disp_max'])
    disp_min_right = read_disp(cfg['input']['disp_min_right'])
    disp_max_right = read_disp(cfg['input']['disp_max_right'])

    # Run the Pandora pipeline
    left, right = run(pandora_machine, img_left, img_right, disp_min, disp_max, cfg['pipeline'], disp_min_right,
                      disp_max_right)

    # Save the left and right DataArray in tiff files
    common.save_results(left, right, output)

    # Save the configuration
    common.save_config(output, cfg)



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains classes and functions associated to the cost volume optimization step.
"""

import logging
import sys
from abc import ABCMeta, abstractmethod
from typing import Dict

import xarray as xr


class AbstractOptimization():
    """
    Abstract Optimizationinput class
    """
    __metaclass__ = ABCMeta

    optimization_methods_avail = {}

    def __new__(cls, **cfg: Dict[str, dict]) -> object:
        """
        Return the plugin associated with the optimization_method given in the configuration

        :param cfg: configuration {'optimization_method': value}
        :type cfg: dictionary
        """
        if cls is AbstractOptimization:
            if isinstance(cfg['optimization_method'], str):
                try:
                    return super(AbstractOptimization, cls).__new__(
                        cls.optimization_methods_avail[cfg['optimization_method']])
                except KeyError:
                    logging.error('No optimization method named % supported', cfg['optimization_method'])
                    sys.exit(1)
            else:
                if isinstance(cfg['optimization_method'], unicode): # pylint: disable=undefined-variable
                    # creating a plugin from registered short name given as unicode (py2 & 3 compatibility)
                    try:
                        return super(AbstractOptimization, cls).__new__(
                            cls.optimization_methods_avail[cfg['optimization_method'].encode('utf-8')])
                    except KeyError:
                        logging.error('No optimization method named % supported', cfg['optimization_method'])
                        sys.exit(1)
        else:
            return super(AbstractOptimization, cls).__new__(cls)

    @classmethod
    def register_subclass(cls, short_name: str):
        """
        Allows to register the subclass with its short name

        :param short_name: the subclass to be registered
        :type short_name: string
        """

        def decorator(subclass):
            """
            Registers the subclass in the available methods

            :param subclass: the subclass to be registered
            :type subclass: object
            """
            cls.optimization_methods_avail[short_name] = subclass
            return subclass

        return decorator

    @abstractmethod
    def desc(self) -> None:
        """
        Describes the optimization method
        :return: None
        """
        print('Optimization method description')

    @abstractmethod
    def optimize_cv(self, cv: xr.Dataset, img_left: xr.Dataset, img_right: xr.Dataset) -> xr.Dataset:
        """
        Optimizes the cost volume

        :param cv: the cost volume dataset with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :type cv: xarray.Dataset
        :param img_left: left Dataset image
        :type img_left: xarray.DataArray
        :param img_right: right Dataset image
        :type img_right: xarray.DataArray
        :return: the cost volume dataset with the data variables:

                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        :rtype: xarray.Dataset
        """



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to ZNCC method used in the cost volume measure step.
"""

from typing import Dict, Union

import numpy as np
import xarray as xr
from json_checker import Checker, And

from pandora.img_tools import shift_right_img, compute_mean_raster, compute_std_raster
from pandora.matching_cost import matching_cost


@matching_cost.AbstractMatchingCost.register_subclass('zncc')
class Zncc(matching_cost.AbstractMatchingCost):
    """
    Zero mean normalized cross correlation
    Zncc class allows to compute the cost volume
    """
    # Default configuration, do not change these values
    _WINDOW_SIZE = 5
    _SUBPIX = 1

    def __init__(self, **cfg: Union[str, int]) -> None:
        """
        :param cfg: optional configuration,  {'window_size': value, 'subpix': value}
        :type cfg: dictionary
        :return: None
        """
        self.cfg = self.check_conf(**cfg)
        self._window_size = self.cfg['window_size']
        self._subpix = self.cfg['subpix']

    def check_conf(self, **cfg: Union[str, int]) -> Dict[str, Union[str, int]]:
        """
        Add default values to the dictionary if there are missing elements and check if the dictionary is correct

        :param cfg: matching cost configuration
        :type cfg: dict
        :return cfg: matching cost configuration updated
        :rtype: dict
        """
        # Give the default value if the required element is not in the configuration
        if 'window_size' not in cfg:
            cfg['window_size'] = self._WINDOW_SIZE
        if 'subpix' not in cfg:
            cfg['subpix'] = self._SUBPIX

        schema = {
            'matching_cost_method': And(str, lambda input: 'zncc'),
            'window_size': And(int, lambda input: input > 0 and (input % 2) != 0),
            'subpix': And(int, lambda input: input in (1, 2, 4))
        }

        checker = Checker(schema)
        checker.validate(cfg)
        return cfg

    def desc(self) -> None:
        """
        Describes the matching cost method
        :return: None
        """
        print('zncc similarity measure')

    def compute_cost_volume(self, img_left: xr.Dataset, img_right: xr.Dataset, disp_min: int,
                            disp_max: int) -> xr.Dataset:
        """
        Computes the cost volume for a pair of images

        :param img_left: left Dataset image
        :type img_left:
            xarray.Dataset containing :
                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :param img_right: right Dataset image
        :type img_right:
            xarray.Dataset containing :
                - im : 2D (row, col) xarray.DataArray
                - msk : 2D (row, col) xarray.DataArray
        :param disp_min: minimum disparity
        :type disp_min: int
        :param disp_max: maximum disparity
        :type disp_max: int
        :return: the cost volume dataset
        :rtype:
            xarray.Dataset, with the data variables:
                - cost_volume 3D xarray.DataArray (row, col, disp)
                - confidence_measure 3D xarray.DataArray (row, col, indicator)
        """
        # Contains the shifted right images
        img_right_shift = shift_right_img(img_right, self._subpix)

        # Computes the standard deviation raster for the whole images
        # The standard deviation raster is truncated for points that are not calculable
        img_left_std = compute_std_raster(img_left, self._window_size)
        img_right_std = []
        for i, img in enumerate(img_right_shift):  # pylint: disable=unused-variable
            img_right_std.append(compute_std_raster(img, self._window_size))

        # Computes the mean raster for the whole images
        # The standard mean raster is truncated for points that are not calculable
        img_left_mean = compute_mean_raster(img_left, self._window_size)
        img_right_mean = []
        for i, img in enumerate(img_right_shift):
            img_right_mean.append(compute_mean_raster(img, self._window_size))

        # Maximal cost of the cost volume with zncc measure
        cmax = 1

        # Cost volume metadata
        offset_row_col = int((self._window_size - 1) / 2)
        metadata = {'measure': 'zncc', 'subpixel': self._subpix,
                    'offset_row_col': offset_row_col, 'window_size': self._window_size,
                    'type_measure': 'max', 'cmax': cmax}

        # Disparity range
        if self._subpix == 1:
            disparity_range = np.arange(disp_min, disp_max + 1)
        else:
            disparity_range = np.arange(disp_min, disp_max, step=1 / float(self._subpix))
            disparity_range = np.append(disparity_range, [disp_max])

        # Allocate the numpy cost volume cv = (disp, col, row), for efficient memory management
        cv = np.zeros((len(disparity_range), img_left['im'].shape[1],
                       img_right['im'].shape[0] ), dtype=np.float32)
        cv += np.nan

        # If offset, do not consider border position for cost computation
        if offset_row_col != 0:
            cv_crop= cv[:, offset_row_col:-offset_row_col, offset_row_col: -offset_row_col]
        else:
            cv_crop = cv

        # Computes the matching cost
        for disp in disparity_range:
            i_right = int((disp % 1) * self._subpix)
            point_p, point_q = self.point_interval(img_left, img_right_shift[i_right], disp)
            dsp = int((disp - disp_min) * self._subpix)

            # Point interval in the left standard deviation image
            # -  (win_radius * 2) because img_std is truncated for points that are not calculable
            p_std = (point_p[0], point_p[1] - (int(self._window_size / 2) * 2))
            # Point interval in the right standard deviation image
            q_std = (point_q[0], point_q[1] - (int(self._window_size / 2) * 2))

            # Compute the normalized summation of the product of intensities
            zncc_ = img_left['im'].data[:, point_p[0]:point_p[1]] * img_right_shift[i_right]['im'].data[:,
                                                                    point_q[0]:point_q[1]]
            zncc_ = xr.Dataset({'im': (['row', 'col'], zncc_)},
                               coords={'row': np.arange(zncc_.shape[0]), 'col': np.arange(zncc_.shape[1])})
            zncc_ = compute_mean_raster(zncc_, self._window_size)
            # Subtracting  the  local mean  value  of  intensities
            zncc_ -= (img_left_mean[:, p_std[0]:p_std[1]] * img_right_mean[i_right][:, q_std[0]:q_std[1]])

            # Divide by the standard deviation of the intensities of the images :
            # If the standard deviation of the intensities of the images is greater than 0
            divide_standard = np.multiply(img_left_std[:, p_std[0]:p_std[1]],
                                          img_right_std[i_right][:, q_std[0]:q_std[1]])
            valid = np.where(divide_standard > 0)
            zncc_[valid] /= divide_standard[valid]
            # Else, zncc = 0
            zncc_[np.where(divide_standard <= 0)] = 0

            # Places the result in the cost_volume
            cv_crop[dsp, point_p[0]:p_std[1], :] = np.swapaxes(zncc_, 0, 1)

        # Create the xarray.DataSet that will contain the cost_volume of dimensions (row, col, disp)
        cv = self.allocate_costvolume(img_left, self._subpix, disp_min, disp_max, self._window_size, metadata,
                                      np.swapaxes(cv, 0, 2))

        return cv



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to raster images.
"""

import logging
import warnings
from typing import List, Union, Tuple

import numpy as np
import rasterio
import xarray as xr
from scipy.ndimage.interpolation import zoom
from skimage.transform.pyramids import pyramid_gaussian

import pandora.constants as cst

def rasterio_open(*args: str, **kwargs: Union[int, str, None]) -> rasterio.io.DatasetReader:
    """
    rasterio.open wrapper to silence UserWarning like NotGeoreferencedWarning.

    (see https://rasterio.readthedocs.io/en/latest/api/rasterio.errors.html)

    :param args: args to be given to rasterio.open method
    :type args: str
    :param kwargs: kwargs to be given to rasterio.open method
    :type kwargs: Union[int, str, None]
    :return: rasterio DatasetReader
    :rtype: rasterio.io.DatasetReader
    """
    # this silence chosen category of warnings only for the following instructions
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=UserWarning)
        return rasterio.open(*args, **kwargs)


def read_img(img: str, no_data: float, mask: str = None, classif: str = None, segm: str = None) ->\
        xr.Dataset:
    """
    Read image and mask, and return the corresponding xarray.DataSet

    :param img: Path to the image
    :type img: string
    :type no_data: no_data value in the image
    :type no_data: float
    :param mask: Path to the mask (optional): 0 value for valid pixels, !=0 value for invalid pixels
    :type mask: string
    :param classif: Path to the classif (optional)
    :type classif: string
    :param segm: Path to the mask (optional)
    :type segm: string
    :return: xarray.DataSet containing the variables :

            - im : 2D (row, col) xarray.DataArray float32
            - msk : 2D (row, col) xarray.DataArray int16, with the convention defined in the configuration file
    :rtype: xarray.DataSet
    """
    img_ds = rasterio_open(img)
    data = img_ds.read(1)

    if np.isnan(no_data):
        no_data_pixels = np.where(np.isnan(data))
    else:
        no_data_pixels = np.where(data == no_data)

    # We accept nan values as no data on input image but to not disturb cost volume processing as stereo computation
    # step,nan as no_data must be converted. We choose -9999 (can be another value). No_data position aren't erased
    # because stored in 'msk'
    if no_data_pixels[0].size != 0 and np.isnan(no_data):
        data[no_data_pixels] = -9999
        no_data = -9999

    dataset = xr.Dataset({'im': (['row', 'col'], data.astype(np.float32))},
                         coords={'row': np.arange(data.shape[0]),
                                 'col': np.arange(data.shape[1])})
    # Add image conf to the image dataset
    dataset.attrs = {'no_data_img': no_data,
                     'valid_pixels': 0, # arbitrary default value
                     'no_data_mask': 1} # arbitrary default value

    if classif is not None:
        input_classif = rasterio_open(classif).read(1)
        dataset['classif'] = xr.DataArray(np.full((data.shape[0], data.shape[1]), 0).astype(np.int16),
                                          dims=['row', 'col'])
        dataset['classif'].data = input_classif

    if segm is not None:
        input_segm = rasterio_open(segm).read(1)
        dataset['segm'] = xr.DataArray(np.full((data.shape[0], data.shape[1]), 0).astype(np.int16),
                                       dims=['row', 'col'])
        dataset['segm'].data = input_segm

    # If there is no mask, and no data in the images, do not create the mask to minimize calculation time
    if mask is None and no_data_pixels[0].size == 0:
        return dataset

    # Allocate the internal mask (!= input_mask)
    # Mask convention:
    # value : meaning
    # dataset.attrs['valid_pixels'] : a valid pixel
    # dataset.attrs['no_data_mask'] : a no_data_pixel
    # other value : an invalid_pixel
    dataset['msk'] = xr.DataArray(np.full((data.shape[0], data.shape[1]),
                                          dataset.attrs['valid_pixels']).astype(np.int16), dims=['row', 'col'])

    # Mask invalid pixels if needed
    # convention: input_mask contains information to identify valid / invalid pixels.
    # Value == 0 on input_mask represents a valid pixel
    # Value != 0 on input_mask represents an invalid pixel
    if mask is not None:
        input_mask = rasterio_open(mask).read(1)
        # Masks invalid pixels
        # All pixels that are not valid_pixels, on the input mask, are considered as invalid pixels
        dataset['msk'].data[np.where(input_mask > 0)] = dataset.attrs['valid_pixels'] + \
                                                        dataset.attrs['no_data_mask'] + 1

    # Masks no_data pixels
    # If a pixel is invalid due to the input mask, and it is also no_data, then the value of this pixel in the
    # generated mask will be = no_data
    dataset['msk'].data[no_data_pixels] = int(dataset.attrs['no_data_mask'])

    return dataset


def prepare_pyramid(img_left: xr.Dataset, img_right: xr.Dataset, num_scales: int, scale_factor: int) -> \
        Tuple[List[xr.Dataset], List[xr.Dataset]]:
    """
    Return a List with the datasets at the different scales

    :param img_left: left Dataset image containing :

            - im : 2D (row, col) xarray.DataArray
    :type img_left: xarray.Dataset
    :param img_right: right Dataset containing :

            - im : 2D (row, col) xarray.DataArray
    :type img_right: xarray.Dataset
    :param num_scales: number of scales
    :type num_scales: int
    :param scale_factor: factor by which downsample the images
    :type scale_factor: int
    :return: a List that contains the different scaled datasets
    :rtype: List of xarray.Dataset
    """
    # Fill no data values in image with an interpolation. If no mask was given, create all valid masks
    img_left_fill, msk_left_fill = fill_nodata_image(img_left)
    img_right_fill, msk_right_fill = fill_nodata_image(img_right)

    # Create image pyramids
    images_left = list(pyramid_gaussian(img_left_fill, max_layer=num_scales - 1, downscale=scale_factor,
                                        sigma=1.2, order=1, mode='reflect', cval=0))
    images_right = list(pyramid_gaussian(img_right_fill, max_layer=num_scales - 1, downscale=scale_factor,
                                         sigma=1.2, order=1, mode='reflect', cval=0))
    # Create mask pyramids
    masks_left = masks_pyramid(msk_left_fill, scale_factor, num_scales)
    masks_right = masks_pyramid(msk_right_fill, scale_factor, num_scales)

    # Create dataset pyramids
    pyramid_left = convert_pyramid_to_dataset(img_left, images_left, masks_left)
    pyramid_right = convert_pyramid_to_dataset(img_right, images_right, masks_right)

    # The pyramid is intended to be from coarse to original size, so we inverse its order.
    return pyramid_left[::-1], pyramid_right[::-1]


def fill_nodata_image(dataset: xr.Dataset) -> Tuple[np.ndarray, np.ndarray]:
    """
        Interpolate no data values in image. If no mask was given, create all valid masks

        :param dataset: Dataset image
        :type dataset: xarray.Dataset containing :

            - im : 2D (row, col) xarray.DataArray
        :return: a Tuple that contains the filled image and mask
        :rtype: Tuple of np.ndarray
        """
    if 'msk' in dataset:
        img, msk = interpolate_nodata_sgm(dataset['im'].data, dataset['msk'].data)
    else:
        msk = np.full((dataset['im'].data.shape[0], dataset['im'].data.shape[1]), int(dataset.attrs['valid_pixels']))
        img = dataset['im'].data
    return img, msk


def interpolate_nodata_sgm(img: np.ndarray, valid: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Interpolation of the input image to resolve invalid (nodata) pixels.
    Interpolate invalid pixels by finding the nearest correct pixels in 8 different directions
    and use the median of their disparities.

    HIRSCHMULLER, Heiko. Stereo processing by semiglobal matching and mutual information.
    IEEE Transactions on pattern analysis and machine intelligence, 2007, vol. 30, no 2, p. 328-341.

    :param img: input image
    :type img: 2D np.array (row, col)
    :param valid: validity mask
    :type valid: 2D np.array (row, col)
    :return: the interpolate input image, with the validity mask update :

        - If out & PANDORA_MSK_PIXEL_FILLED_NODATA != 0 : Invalid pixel : filled nodata pixel
    :rtype: tuple(2D np.array (row, col), 2D np.array (row, col))
    """
    # Output disparity map and validity mask
    out_img = np.copy(img)
    out_val = np.copy(valid)

    ncol, nrow = img.shape

    # 8 directions : [row, y]
    dirs = np.array([[0, 1], [-1, 1], [-1, 0], [-1, -1], [0, -1], [1, -1], [1, 0], [1, 1]])
    # Maximum path length
    max_path_length = max(nrow, ncol)

    ncol, nrow = img.shape
    for col in range(ncol):
        for row in range(nrow):
            # Mismatched
            if valid[col, row] & cst.PANDORA_MSK_PIXEL_INVALID != 0:
                # For each directions
                valid_neighbors = np.zeros(8, dtype=np.float32)
                for dir_ind in range(8):
                    # Find the first valid pixel in the current path
                    tmp_row = row
                    tmp_col = col
                    for i in range(max_path_length): #pylint: disable=unused-variable
                        tmp_row += dirs[dir_ind][0]
                        tmp_col += dirs[dir_ind][1]

                        # Edge of the image reached: there is no valid pixel in the current path
                        if (tmp_col < 0) | (tmp_col >= ncol) | (tmp_row < 0) | (tmp_row >= nrow):
                            valid_neighbors[dir_ind] = np.nan
                            break

                        # First valid pixel
                        if (valid[tmp_col, tmp_row] & cst.PANDORA_MSK_PIXEL_INVALID) == 0:
                            valid_neighbors[dir_ind] = img[tmp_col, tmp_row]
                            break

                # Median of the 8 pixels
                out_img[col, row] = np.nanmedian(valid_neighbors)
                # Update the validity mask : Information : filled nodata pixel
                out_val[col, row] = cst.PANDORA_MSK_PIXEL_FILLED_NODATA

    return out_img, out_val


def masks_pyramid(msk: np.ndarray, scale_factor: int, num_scales: int) -> List[np.ndarray]:
    """
    Return a List with the downsampled masks for each scale

    :param msk: full resolution mask
    :type msk: np.ndarray
    :param scale_factor: scale factor
    :type scale_factor: int
    :param num_scales: number of scales
    :type num_scales: int
    :return: a List that contains the different scaled masks
    :rtype: List of np.ndarray
    """
    msk_pyramid = []
    # Add the full resolution mask
    msk_pyramid.append(msk)
    tmp_msk = msk
    for scale in range(num_scales - 1): #pylint: disable=unused-variable
        # Decimate in the two axis
        tmp_msk = tmp_msk[::scale_factor, ::scale_factor]
        msk_pyramid.append(tmp_msk)
    return msk_pyramid


def convert_pyramid_to_dataset(img_orig: xr.Dataset, images: List[np.ndarray], masks: List[np.ndarray]) \
        -> List[xr.Dataset]:
    """
    Return a List with the datasets at the different scales

    :param img_left: left Dataset image containing :

        - im : 2D (row, col) xarray.DataArray
    :type img_left: xarray.Dataset
    :param img_right: right Dataset image containing :

        - im : 2D (row, col) xarray.DataArray
    :type img_right: xarray.Dataset
    :param num_scales: number of scales
    :type num_scales: int
    :param scale_factor: factor by which downsample the images
    :type scale_factor: int
    :return: a List that contains the different scaled datasets
    :rtype: List of xarray.Dataset
    """


    pyramid = []
    for index, image in enumerate(images):
        # The full resolution image (first in list) has to be the original image and mask
        if index == 0:
            pyramid.append(img_orig)
            continue

        # Creating new dataset
        dataset = xr.Dataset({'im': (['row', 'col'], image.astype(np.float32))},
                             coords={'row': np.arange(image.shape[0]),
                                     'col': np.arange(image.shape[1])})

        # Allocate the mask
        dataset['msk'] = xr.DataArray(np.full((image.shape[0], image.shape[1]), masks[index].astype(np.int16)),
                                      dims=['row', 'col'])

        # Add image conf to the image dataset
        dataset.attrs = {'no_data_img': img_orig.attrs['no_data_img'],
                         'valid_pixels': img_orig.attrs['valid_pixels'],
                         'no_data_mask': img_orig.attrs['no_data_mask']}
        pyramid.append(dataset)

    return pyramid


def shift_right_img(img_right: xr.Dataset, subpix: int) -> List[xr.Dataset]:
    """
    Return an array that contains the shifted right images

    :param img_right: right Dataset image containing :

        - im : 2D (row, col) xarray.DataArray
    :type img_right: xarray.Dataset
    :param subpix: subpixel precision = (1 or 2 or 4)
    :type subpix: int
    :return: an array that contains the shifted right images
    :rtype: array of xarray.Dataset
    """
    img_right_shift = [img_right]
    ny_, nx_ = img_right['im'].shape

    # zoom factor = (number of columns with zoom / number of original columns)
    if subpix == 2:
        # Shift the right image for subpixel precision 0.5
        data = zoom(img_right['im'].data, (1, (nx_ * 4 - 3) / float(nx_)), order=1)[:, 2::4]
        col = np.arange(img_right.coords['col'][0] + 0.5, img_right.coords['col'][-1], step=1)
        img_right_shift.append(xr.Dataset({'im': (['row', 'col'], data)},
                                          coords={'row': np.arange(ny_), 'col': col}))

    if subpix == 4:
        # Shift the right image for subpixel precision 0.25
        data = zoom(img_right['im'].data, (1, (nx_ * 4 - 3) / float(nx_)), order=1)[:, 1::4]
        col = np.arange(img_right.coords['col'][0] + 0.25, img_right.coords['col'][-1], step=1)
        img_right_shift.append(xr.Dataset({'im': (['row', 'col'], data)},
                                          coords={'row': np.arange(ny_), 'col': col}))

        # Shift the right image for subpixel precision 0.5
        data = zoom(img_right['im'].data, (1, (nx_ * 4 - 3) / float(nx_)), order=1)[:, 2::4]
        col = np.arange(img_right.coords['col'][0] + 0.5, img_right.coords['col'][-1], step=1)
        img_right_shift.append(xr.Dataset({'im': (['row', 'col'], data)},
                                          coords={'row': np.arange(ny_), 'col': col}))

        # Shift the right image for subpixel precision 0.75
        data = zoom(img_right['im'].data, (1, (nx_ * 4 - 3) / float(nx_)), order=1)[:, 3::4]
        col = np.arange(img_right.coords['col'][0] + 0.75, img_right.coords['col'][-1], step=1)
        img_right_shift.append(xr.Dataset({'im': (['row', 'col'], data)},
                                          coords={'row': np.arange(ny_), 'col': col}))
    return img_right_shift


def check_inside_image(img: xr.Dataset, row: int, col: int) -> bool:
    """
    Check if the coordinates row,col are inside the image

    :param img: Dataset image containing :

            - im : 2D (row, col) xarray.DataArray
    :type img: xarray.Dataset
    :param row: row coordinates
    :type row: int
    :param col: column coordinates
    :type col: int
    :return: a boolean
    :rtype: boolean
    """
    nx_, ny_ = img['im'].shape
    return 0 <= row < nx_ and 0 <= col < ny_


def census_transform(image: xr.Dataset, window_size: int) -> xr.Dataset:
    """
    Generates the census transformed image from an image

    :param image: Dataset image containing the image im : 2D (row, col) xarray.Dataset
    :type image: xarray.Dataset
    :param window_size: Census window size
    :type window_size: int
    :return: Dataset census transformed uint32 containing the transformed image im: 2D (row, col) xarray.DataArray \
    uint32
    :rtype: xarray.Dataset
    """
    ny_, nx_ = image['im'].shape
    border = int((window_size - 1) / 2)

    # Create a sliding window of using as_strided function : this function create a new a view (by manipulating data
    #  pointer) of the image array with a different shape. The new view pointing to the same memory block as
    # image so it does not consume any additional memory.
    str_row, str_col = image['im'].data.strides
    shape_windows = (ny_ - (window_size - 1), nx_ - (window_size - 1), window_size, window_size)
    strides_windows = (str_row, str_col, str_row, str_col)
    windows = np.lib.stride_tricks.as_strided(image['im'].data, shape_windows, strides_windows, writeable=False)

    # Pixels inside the image which can be centers of windows
    central_pixels = image['im'].data[border:-border, border:-border]

    # Allocate the census mask
    census = np.zeros((ny_ - (window_size - 1), nx_ - (window_size - 1)), dtype='uint32')

    shift = (window_size * window_size) - 1
    for row in range(window_size):
        for col in range(window_size):
            # Computes the difference and shift the result
            census[:, :] += ((windows[:, :, row, col] > central_pixels[:, :]) << shift).astype(np.uint32)
            shift -= 1

    census = xr.Dataset({'im': (['row', 'col'], census)},
                        coords={'row': np.arange(border, ny_ - border),
                                'col': np.arange(border, nx_ - border)})

    return census


def compute_mean_raster(img: xr.Dataset, win_size: int) -> np.ndarray:
    """
    Compute the mean within a sliding window for the whole image

    :param img: Dataset image containing :

            - im : 2D (row, col) xarray.DataArray
    :type img: xarray.Dataset
    :param win_size: window size
    :type win_size: int
    :return: mean raster
    :rtype: numpy array
    """
    ny_, nx_ = img['im'].shape

    # Example with win_size = 3 and the input :
    #           10 | 5  | 3
    #            2 | 10 | 5
    #            5 | 3  | 1

    # Compute the cumulative sum of the elements along the column axis
    r_mean = np.r_[np.zeros((1, nx_)), img['im']]
    # r_mean :   0 | 0  | 0
    #           10 | 5  | 3
    #            2 | 10 | 5
    #            5 | 3  | 1
    r_mean = np.cumsum(r_mean, axis=0)
    # r_mean :   0 | 0  | 0
    #           10 | 5  | 3
    #           12 | 15 | 8
    #           17 | 18 | 9
    r_mean = r_mean[win_size:, :] - r_mean[:-win_size, :]
    # r_mean :  17 | 18 | 9

    # Compute the cumulative sum of the elements along the row axis
    r_mean = np.c_[np.zeros(ny_ - (win_size - 1)), r_mean]
    # r_mean :   0 | 17 | 18 | 9
    r_mean = np.cumsum(r_mean, axis=1)
    # r_mean :   0 | 17 | 35 | 44
    r_mean = r_mean[:, win_size:] - r_mean[:, :-win_size]
    # r_mean : 44
    return r_mean / float(win_size * win_size)


def compute_mean_patch(img: xr.Dataset, row: int, col: int, win_size: int) -> np.ndarray:
    """
    Compute the mean within a window centered at position row,col

    :param img: Dataset image containing :

            - im : 2D (row, col) xarray.DataArray
    :type img: xarray.Dataset
    :param row: row coordinates
    :type row: int
    :param col: column coordinates
    :type col: int
    :param win_size: window size
    :type win_size: int
    :return: mean
    :rtype: float
    """
    begin_window = (row - int(win_size / 2), col - int(win_size / 2))
    end_window = (row + int(win_size / 2), col + int(win_size / 2))

    # Check if the window is inside the image, and compute the mean
    if check_inside_image(img, begin_window[0], begin_window[1]) and \
            check_inside_image(img, end_window[0], end_window[1]):
        return np.mean(img['im'][begin_window[1]:end_window[1] + 1, begin_window[0]:end_window[0] + 1],
                       dtype=np.float32)

    logging.error('The window is outside the image')
    raise IndexError


def compute_std_raster(img: xr.Dataset, win_size: int) -> np.ndarray:
    """
    Compute the standard deviation within a sliding window for the whole image
    with the formula : std = sqrt( E[row^2] - E[row]^2 )

    :param img: Dataset image containing :

            - im : 2D (row, col) xarray.DataArray
    :type img: xarray.Dataset
    :param win_size: window size
    :type win_size: int
    :return: std raster
    :rtype: numpy array
    """
    # Computes E[row]
    mean_ = compute_mean_raster(img, win_size)

    # Computes E[row^2]
    raster_power_two = xr.Dataset({'im': (['row', 'col'], img['im'].data ** 2)},
                                  coords={'row': np.arange(img['im'].shape[0]), 'col': np.arange(img['im'].shape[1])})
    mean_power_two = compute_mean_raster(raster_power_two, win_size)

    # Compute sqrt( E[row^2] - E[row]^2 )
    var = mean_power_two - mean_ ** 2
    # Avoid very small values
    var[np.where(var < (10 ** (-15) * abs(mean_power_two)))] = 0
    return np.sqrt(var)


def read_disp(disparity: Union[None, int, str]) -> Union[None, int, np.ndarray]:
    """
    Read the disparity :
        - if cfg_disp is the path of a disparity grid, read and return the grid (type numpy array)
        - else return the value of cfg_disp

    :param disparity: disparity, or path to the disparity grid
    :type disparity: None, int or str
    :return: the disparity
    :rtype: int or np.ndarray
    """
    if isinstance(disparity, str):
        disp_ = rasterio_open(disparity)
        data_disp = disp_.read(1)
    else:
        data_disp = disparity

    return data_disp



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of PANDORA
#
#     https://github.com/CNES/Pandora_pandora
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions associated to the multiscale step.
"""

import logging
from abc import ABCMeta, abstractmethod
from typing import Tuple, Union

import numpy as np
import xarray as xr

import pandora.constants as cst


class AbstractMultiscale:
    """
    Abstract Multiscale class
    """
    __metaclass__ = ABCMeta

    multiscale_methods_avail = {}

    def __new__(cls, **cfg: Union[str, int]):
        """
        Return the plugin associated with the multiscale method given in the configuration

        :param cfg: configuration {'multiscale_method': value, 'margin': value}
        :type cfg: dictionary
        """
        if cls is AbstractMultiscale:
            if isinstance(cfg['multiscale_method'], str):
                try:
                    return super(AbstractMultiscale, cls).__new__(
                        cls.multiscale_methods_avail[cfg['multiscale_method']])
                except KeyError:
                    logging.error('No multiscale matching method named % supported', cfg['multiscale_method'])
                    raise KeyError
            else:
                if isinstance(cfg['multiscale_method'], unicode):  # pylint:disable=undefined-variable
                    # creating a plugin from registered short name given as unicode (py2 & 3 compatibility)
                    try:
                        return super(AbstractMultiscale, cls).__new__(
                            cls.multiscale_methods_avail[cfg['multiscale_method'].encode('utf-8')])
                    except KeyError:
                        logging.error(
                            'No multiscale matching method named % supported', cfg['multiscale_method'])
                        raise KeyError
        else:
            return super(AbstractMultiscale, cls).__new__(cls)

    @classmethod
    def register_subclass(cls, short_name: str, *args):
        """
        Allows to register the subclass with its short name

        :param short_name: the subclass to be registered
        :type short_name: string
        :param args: allows to register one plugin that contains different methods
        """

        def decorator(subclass):
            """
            Registers the subclass in the available methods

            :param subclass: the subclass to be registered
            :type subclass: object
            """
            cls.multiscale_methods_avail[short_name] = subclass
            for arg in args:
                cls.multiscale_methods_avail[arg] = subclass
            return subclass

        return decorator

    @abstractmethod
    def desc(self):
        """
        Describes the multiscale method
        """
        print('Multiscale matching description')

    @abstractmethod
    def disparity_range(self, disp: xr.Dataset, disp_min: int, disp_max: int) -> \
            Tuple[np.array, np.array]:
        """
        Disparity range computation by seeking the max and min values in the window.
        Unvalid disparities are given the full disparity range

        :param disp: the disparity dataset
        :type disp: xarray.Dataset with the data variables :
                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray(row, col, indicator)
        :param disp_min: absolute min disparity
        :type disp_min: int
        :param disp_max: absolute max disparity
        :type disp_max: int
        :return: Two np.darray :
                - disp_min_range : minimum disparity value for all pixels :
                - disp_max_range : maximum disparity value for all pixels.

        :rtype: tuple (np.ndarray, np.ndarray)
        """

    @staticmethod
    def mask_invalid_disparities(disp: xr.Dataset) -> np.ndarray:
        """
        Return a copied disparity map with all invalid disparities set to Nan

        :param disp: the disparity dataset
        :type disp: xarray.Dataset with the data variables :
                - disparity_map 2D xarray.DataArray (row, col)
                - confidence_measure 3D xarray.DataArray(row, col, indicator)

        :return: np.darray :
                - filtered_disp_map : disparity map with invalid values set to Nzn

        :rtype: tuple (np.ndarray, np.ndarray)
        """
        filtered_disp_map = disp['disparity_map'].data.copy()

        # Set all invalid disparities to nan
        for idxes, val in np.ndenumerate(disp['validity_mask'].data):
            if val & cst.PANDORA_MSK_PIXEL_INVALID != 0:
                filtered_disp_map[idxes] = np.nan

        return filtered_disp_map
